--                                                                    --
--  procedure                       Copyright (c)  Dmitry A. Kazakov  --
--     Gtk.Abstract_Browser                        Luebeck            --
--  Implementation                                 Autumn, 2007       --
--                                                                    --
--                                Last revision :  09:24 09 Apr 2010  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.Exceptions;        use Ada.Exceptions;
with Ada.IO_Exceptions;     use Ada.IO_Exceptions;
with Gdk.Color;             use Gdk.Color;
with Gdk.Rectangle;         use Gdk.Rectangle;
with Gdk.Types.Keysyms;     use Gdk.Types.Keysyms;
with GLib.Properties;       use GLib.Properties;
with GLib.Messages;         use GLib.Messages;
with GLib.Values;           use GLib.Values;
with GLib.Values.Handling;  use GLib.Values.Handling;
with GLib.Unicode;          use GLib.Unicode;
with Gtk.Enums;             use Gtk.Enums;
with Gtk.Widget;            use Gtk.Widget;
with Gtk.Scrolled_Window;   use Gtk.Scrolled_Window;
with Gtk.Stock;             use Gtk.Stock;
with Gtk.Style;             use Gtk.Style;

with Ada.Unchecked_Deallocation;
with Gdk.Drawable;
with Gdk.Window;
with GtkAda.Types;
with Interfaces.C.Strings;

package body Gtk.Abstract_Browser is

   Cell_BG_GDK : constant Property_Address :=
                             Build ("cell-background-gdk");
   Cell_BG_Set : constant Property_Boolean :=
                             Build ("cell-background-set");
   BG_GDK : constant Property_Address := Build ("background-gdk");
   BG_Set : constant Property_Boolean := Build ("background-set");
   FG_GDK : constant Property_Address := Build ("foreground-gdk");
   FG_Set : constant Property_Boolean := Build ("foreground-set");

   function Where (Name : String) return String is
   begin
      return " in Gtk.Abstract_Browser." & Name;
   end Where;

   Abstract_Directory_Class : GObject_Class := Uninitialized_Class;
   Directory_Items_Class    : GObject_Class := Uninitialized_Class;

   Abstract_Directory_Signals : constant GtkAda.Types.Chars_Ptr_Array :=
      (  0 => Interfaces.C.Strings.New_String ("read-error"),
         1 => Interfaces.C.Strings.New_String ("rewind-error"),
         2 => Interfaces.C.Strings.New_String ("refreshed")
      );
   Directory_Items_Signals : constant GtkAda.Types.Chars_Ptr_Array :=
      (  0 => Interfaces.C.Strings.New_String ("directory-changed"),
         1 => Interfaces.C.Strings.New_String ("selection-changed")
      );

   procedure Activated
             (  Widget : access Gtk_Directory_Items_View_Record;
                Index  : Positive
             )  is
   begin
      if Is_Directory (Widget, Index) then
         Set_Current_Directory
         (  Widget.Directories,
            Get_Path (Widget, Index)
         );
      elsif Is_Editable (Widget) then
         --
         -- When  the  widget has editable names, then activation starts
         -- editing.
         --
         declare
            Row    : Gtk_Tree_Iter := Get_Iter (Widget, Index);
            Column : Positive      := Get_Column (Widget, Index);
            Path   : Gtk_Tree_Path;
         begin
            if Row = Null_Iter then
               Log
               (  GtkAda_Contributions_Domain,
                  Log_Level_Critical,
                  (  "Invisible item activated"
                  &  Where ("Activated")
               )  );
            else
               To_Columned (Widget.Columns, Row, Column);
               Path := Get_Path (Widget.Columns, Row);
               Set_Cursor_On_Cell
               (  Tree_View     => Widget,
                  Path          => Path,
                  Start_Editing => True,
                  Focus_Column  =>
                     Get_Column (Widget, GInt (Column) - 1),
                  Focus_Cell    =>
                     Widget.Name_Renderers (Column).all'Access
               );
               Path_Free (Path);
            end if;
         end;
      end if;
   end Activated;

   procedure Add_Folder
             (  Store     : access Gtk_Abstract_Directory_Record'Class;
                Row       : in out Gtk_Tree_Iter;
                Directory : Item_Path;
                Expanded  : Boolean;
                Updated   : out Boolean
             )  is
      Item    : Gtk_Tree_Iter;
      Path    : Gtk_Tree_Path;
      Recurse : Natural := 0;
   begin
      if Row = Null_Iter then
         Item := Get_Iter_First (Store.Tree);
      else
         if Get_Int (Store.Tree, Row, 2) not in Cached_Directory then
            Updated := False;
            return;
         end if;
         Item := Children (Store.Tree, Row);
      end if;
      if Item /= Null_Iter then
         -- This directory is already cached
         if Expanded then
            Expand_Folder (Store, Row, Directory);
         end if;
         Updated := False;
         return;
      end if;
      -- Prepare to read the directory items out
      begin
         declare
            Data : Directory_Item := Rewind (Store, Directory);
         begin
            case Data.Policy is
               when Cache_Never =>
                  if Row /= Null_Iter then
                     Set_Item (Store, Row, Data);
                  end if;
                  Updated := True;
                  return;
               when Cache_Expanded =>
                  Data.Directory := True;
                  if Row /= Null_Iter then
                     Set_Item (Store, Row, Data);
                  end if;
               when Cache_Ahead =>
                  null;
            end case;
         end;
      exception
         when Error : Data_Error => -- Removing the directory from cache
            Remove (Store.Tree, Row);
            Row := Null_Iter;
            Rewind_Error
            (  Store,
               Exception_Message (Error),
               Directory
            );
            Updated := True;
            return;
         when Error : others =>
            Log
            (  GtkAda_Contributions_Domain,
               Log_Level_Critical,
               (  "Rewind error:"
               &  Exception_Information (Error)
               &  Where ("Add_Folder")
            )  );
            Remove (Store.Tree, Row);
            Row := Null_Iter;
            Rewind_Error
            (  Store,
               Exception_Message (Error),
               Directory
            );
            Updated := True;
            return;
      end;
      --
      -- Here  we  will modify the store and we don't know if that would
      -- have an effect on the iterators of. For this reason we save the
      -- iterator as a path, which won't change and restore the iterator
      -- after each suspicious operation.
      --
      if Row /= Null_Iter then
         Path := Get_Path (Store.Tree, Row);
      end if;
      loop
         declare
            Data : Directory_Item renames Read (Store);
         begin
            Item := Add_Item (Store, Row, Directory, Data, False);
            if (  Item /= Null_Iter
               and then
                  Data.Directory
               and then
                  Data.Policy in Cache_Expanded..Cache_Ahead
               )
            then
               Recurse := Recurse + 1;
            end if;
            if Path = null then
               Row := Null_Iter;
            else
               Row := Get_Iter (Store.Tree, Path);
            end if;
         end;
      end loop;
   exception
      when End_Error =>
         --
         -- No  more items. We go though the directory items in order to
         -- check if some of them are themselves directories  needed  to
         -- be cached.
         --
         for Index in reverse 0..N_Children (Store.Tree, Row) - 1 loop
            exit when Recurse = 0;
            Row := Nth_Child (Store.Tree, Row, Index);
            exit when Row = Null_Iter;
            case Get_Int (Store.Tree, Row, 2) is
               when Cached_Ahead_Directory =>
                  Add_Folder
                  (  Store,
                     Row,
                     Get_Path
                     (  Store,
                        Directory,
                        Item_Name (Get_String (Store.Tree, Row, 1))
                     ),
                     True,
                     Updated
                  );
                  Recurse := Recurse - 1;
               when Cached_Expanded_Directory =>
                  if Expanded then
                     Add_Folder
                     (  Store,
                        Row,
                        Get_Path
                        (  Store,
                           Directory,
                           Item_Name (Get_String (Store.Tree, Row, 1))
                        ),
                        False,
                        Updated
                     );
                  end if;
                  Recurse := Recurse - 1;
               when others =>
                  null;
            end case;
            if Path = null then
               Row := Null_Iter;
            else
               Row := Get_Iter (Store.Tree, Path);
            end if;
         end loop;
         Path_Free (Path);
         Updated := True;
      when Error : Data_Error =>
         if Path = null then
            Row := Null_Iter;
         else
            Row := Get_Iter (Store.Tree, Path);
            Path_Free (Path);
            Remove (Store.Tree, Row);
            Row := Null_Iter;
         end if;
         Read_Error
         (  Store,
            Exception_Message (Error),
            Directory
         );
         Updated := True;
      when Error : others =>
         Log
         (  GtkAda_Contributions_Domain,
            Log_Level_Critical,
            (  "Read error: "
            &  Exception_Information (Error)
            &  Where ("Add_Folder")
         )  );
         if Path = null then
            Row := Null_Iter;
         else
            Row := Get_Iter (Store.Tree, Path);
            Path_Free (Path);
            Remove (Store.Tree, Row);
            Row := Null_Iter;
         end if;
         Read_Error
         (  Store,
            Exception_Message (Error),
            Directory
         );
         Updated := True;
   end Add_Folder;

   function Add_Item
            (  Store     : access Gtk_Abstract_Directory_Record'Class;
               Row       : Gtk_Tree_Iter;
               Directory : Item_Path;
               Item      : Directory_Item;
               Update    : Boolean
            )  return Gtk_Tree_Iter is
      Position : constant Integer := Find (Store, Row, Directory, Item);
      Result   : Gtk_Tree_Iter;
   begin
      if Position < 0 then
         Store.Inserting := Store.Inserting + 1;
         Insert (Store.Tree, Result, Row, GInt (-Position) - 1);
         Store.Inserting := Store.Inserting - 1;
         Set_Item (Store, Result, Item);
      else
         Result := Nth_Child (Store.Tree, Row, GInt (Position) - 1);
         if Update then
            Set_Item (Store, Result, Item);
         end if;
      end if;
      return Result;
   end Add_Item;

   procedure Adjust (Object : in out Selection_History) is
   begin
      Object.Last := null;
      Object.This := null;
   end Adjust;

   procedure Cache
             (  Store : access Gtk_Abstract_Directory_Record'Class;
                Item  : Item_Path
             )  is
      Best_Match, Exact_Match : Gtk_Tree_Iter;
   begin
      Cache (Store, Item, Best_Match, Exact_Match);
   end Cache;

   procedure Cache
             (  Store : access Gtk_Abstract_Directory_Record'Class;
                Item        : Item_Path;
                Best_Match  : out Gtk_Tree_Iter;
                Exact_Match : out Gtk_Tree_Iter
             )  is
      Updated : Boolean;
   begin
      if Item'Length = 0 then
         Best_Match  := Null_Iter;
         Exact_Match := Null_Iter;
         return;
      end if;
      begin
         declare
            Directory : constant Item_Path :=
                           Get_Directory (Store, Item);
         begin
            -- Searching in a nested directory
            Cache (Store, Directory, Best_Match, Exact_Match);
            if Exact_Match /= Null_Iter then
               -- The parent was found and its children cached
               if Exact_Match /= Null_Iter then
                  Exact_Match :=
                     Find
                     (  Store,
                        Best_Match,
                        Directory,
                        Get_Name (Store, Item)
                     );
               end if;
            end if;
         end;
      exception
         when Name_Error =>
            -- Searching in the root directory
            Best_Match  := Null_Iter;
            Exact_Match := Null_Iter;
            Add_Folder (Store, Exact_Match, "", False, Updated);
            -- The root-level items have been cached
            Exact_Match :=
               Find
               (  Store,
                  Best_Match,
                  "",
                  Get_Name (Store, Item)
               );
      end;
      if Exact_Match /= Null_Iter then
         if Get_Int (Store.Tree, Exact_Match, 2) in Cached_Directory
         then  -- A directory was found, it is cached
            Add_Folder (Store, Exact_Match, Item, False, Updated);
         end if;
         Best_Match := Exact_Match;
      end if;
   end Cache;

   procedure Changed
             (  Store     : access Gtk_Abstract_Directory_Record'Class;
                Directory : Item_Path
             )  is
      Row : Gtk_Tree_Iter := Find (Store, Directory);
   begin
      if Row /= Null_Iter then
         Refresh (Store, Row);
      end if;
   end Changed;

   procedure Changed
             (  Store : access Gtk_Abstract_Directory_Record'Class
             )  is
      Row : Gtk_Tree_Iter := Null_Iter;
   begin
      Refresh (Store, Row);
   end Changed;

   procedure Change_Selection
             (  Widget : access Gtk_Directory_Items_View_Record;
                Index  : Positive;
                State  : Boolean
             )  is
      Row    : Gtk_Tree_Iter := Get_Iter (Widget, Index);
      Column : Positive      := Get_Column (Widget, Index);
      Path   : Gtk_Tree_Path;
      Area   : Gdk_Rectangle;
   begin
      if Row = Null_Iter then
         Log
         (  GtkAda_Contributions_Domain,
            Log_Level_Critical,
            (  "Non-existing item selected"
            &  Where ("Set_Selection")
         )  );
      else
         Set_Extension (Widget.Markup, Row, 1, State);
         if State then
            Widget.Markup.Selected := Widget.Markup.Selected + 1;
         else
            Widget.Markup.Selected := Widget.Markup.Selected - 1;
         end if;
         --
         -- Forcing rendering the selected column
         --
         To_Columned (Widget.Columns, Row, Column);
         Path := Get_Path (Widget.Columns, Row);
         Get_Cell_Area
         (  Widget,
            Path,
            Get_Column (Widget, GInt (Column)),
            Area
         );
         Path_Free (Path);
         Gdk.Window.Invalidate_Rect
         (  Get_Bin_Window (Widget),
            Area,
            True
         );
      end if;
   end Change_Selection;

   procedure Created
             (  Store     : access Gtk_Abstract_Directory_Record'Class;
                Directory : Item_Path;
                Item      : Directory_Item
             )  is
      Row : Gtk_Tree_Iter := Find (Store, Directory);
   begin
      if Row = Null_Iter then
         Row := Add_Item (Store, Null_Iter, "", Item, False);
      else
         Row := Add_Item (Store, Row, Directory, Item, False);
      end if;
   end Created;

   procedure Deleted
             (  Store : access Gtk_Abstract_Directory_Record'Class;
                Item  : Item_Path
             )  is
      Row : Gtk_Tree_Iter;
   begin
      Row := Find (Store, Item);
      if Row /= Null_Iter then
         Remove (Store.Tree, Row);
      end if;
   end Deleted;

   procedure Deleted
             (  Model : access Gtk_Selection_Store_Record;
                Path  : Gtk_Tree_Path
             )  is
   begin
      Deleted (Gtk_Extension_Store_Record (Model.all)'Access, Path);
      if Model.Changed then
         Model.Changed := False;
         if (  Model.Mode = Selection_Browse
            and then
               Model.Selected = 0
            and then
               Get_Directory_Size (Model.View) > 0
            )
         then
            -- Select another item
            declare
               Current : Natural := Get_Current (Model.View);
            begin
               if Current > 0 then
                  Change_Selection (Model.View, Current, True);
               else
                  Change_Selection (Model.View, 1, True);
               end if;
            end;
         end if;
         Selection_Changed (Model.View);
      end if;
   end Deleted;

   procedure Deleting
             (  Model : access Gtk_Selection_Store_Record;
                Path  : Gtk_Tree_Path;
                Iter  : Gtk_Tree_Iter
             )  is
   begin
      if Get_Boolean (Model, Iter, 3) then
         Model.Selected := Model.Selected - 1;
         Model.Changed  := True;
      end if;
   end Deleting;

   procedure Destroy
             (  Object  : access GObject_Record'Class;
                Browser : Gtk_Directory_Items_View
             )  is
   begin
      Finalize (Browser);
   end Destroy;

   procedure Directory_Activated
             (  Object  : access GObject_Record'Class;
                Params  : GValues;
                Browser : Gtk_Directory_Tree_View
             )  is
      Row : Gtk_Tree_Iter :=
               Get_Iter
               (  Browser.Cache,
                  Convert (Get_Address (Nth (Params, 1)))
               );
   begin
      if Row /= Null_Iter then
         Select_Iter (Get_Selection (Browser), Row);
      end if;
   end Directory_Activated;

   procedure Directory_Changed
             (  Widget : access Gtk_Directory_Items_View_Record
             )  is
   begin
      Directory_Tree_Handlers.Emit_By_Name
      (  Widget,
         "directory-changed"
      );
   end Directory_Changed;

   procedure Directory_Changed
             (  Object  : access GObject_Record'Class;
                Browser : Gtk_Directory_Tree_View
             )  is
      Row   : Gtk_Tree_Iter;
      Model : Gtk_Tree_Model;
      Path  : Gtk_Tree_Path;
   begin
      Get_Selected (Get_Selection (Browser), Model, Row);
      if Row = Null_Iter then
         if Browser.Selection.This = null then
            return;
         end if;
      else
         Path := Get_Path (Browser.Cache, Row);
         if (  Browser.Selection.This /= null
            and then
               Compare (Path, Browser.Selection.This) = 0
            )
         then
            Path_Free (Path);
            return;
         end if;
      end if;
      if Browser.Selection.Last = null then
         Path_Free (Browser.Selection.Last);
      end if;
      Browser.Selection.Last := Browser.Selection.This;
      Browser.Selection.This := Path;
   end Directory_Changed;

   procedure Directory_Collapsed
             (  Object  : access GObject_Record'Class;
                Params  : GValues;
                Browser : Gtk_Directory_Items_View
             )  is
      Row  : Gtk_Tree_Iter;
      Item : Gtk_Tree_Iter;
   begin
      Get_Tree_Iter (Nth (Params, 1), Row);
      if Row = Null_Iter then
         return;
      end if;
      Item := To_Item (Browser.Cache, Row);
      if Row = Null_Iter then
         return;
      end if;
      Item := To_Marked (Browser, Item);
      if (  Item = Null_Iter
         or else
            not Is_In
                (  Browser.Markup,
                   Gtk_Tree_Iter'(Get_Root (Browser.Columns)),
                   Item
         )      )
      then
         return;
      end if;
      --
      -- A   directory   containing  the  currently  selected  one  gets
      -- collapsed. We set the  selection  to  the  collapsed  directory
      -- which in turn chages the current directory.
      --
      Select_Iter (Get_Selection (Browser.Directories), Row);
   end Directory_Collapsed;

   procedure Directory_Expanded
             (  Object  : access GObject_Record'Class;
                Params  : GValues;
                Browser : Gtk_Directory_Tree_View
             )  is
      Updated : Boolean;
      Row     : Gtk_Tree_Iter;
   begin
      Get_Tree_Iter (Nth (Params, 1), Row);
      if Row /= Null_Iter then
         Row := To_Item (Browser.Cache, Row);
         Add_Folder
         (  Browser.Cache,
            Row,
            Get_Path (Browser.Cache, Row),
            True,
            Updated
         );
      end if;
   end Directory_Expanded;

   procedure Directory_Refreshed
             (  Object  : access GObject_Record'Class;
                Params  : GValues;
                Browser : Gtk_Directory_Tree_View
             )  is
      Path : String := Get_String (Nth (Params, 1));
      Row  : Gtk_Tree_Iter;
   begin
      Row := Find (Browser.Cache, Item_Path (Path));
      if Row /= Null_Iter then
         Row := From_Item (Browser.Cache, Row);
         declare
            Last : Gtk_Tree_Iter := Null_Iter;
            This : Gtk_Tree_Iter := Null_Iter;
         begin
            if Browser.Selection.Last /= null then
               Last := Get_Iter (Browser.Cache, Browser.Selection.Last);
            end if;
            if Browser.Selection.This /= null then
               This := Get_Iter (Browser.Cache, Browser.Selection.This);
            end if;
            if This = Row or else Last = Row then
               -- Restoring lost selection
               Unselect_All (Get_Selection (Browser));
               Select_Iter (Get_Selection (Browser), Row);
            end if;
         end;
      end if;
   end Directory_Refreshed;

   procedure Edited_Directory
             (  Object  : access GObject_Record'Class;
                Params  : GValues;
                Browser : Gtk_Directory_Tree_View
             )  is
      Row : Gtk_Tree_Iter :=
               Get_Iter_From_String
               (  Browser.Cache,
                  Get_String (Nth (Params, 1))
               );
   begin
      if Row = Null_Iter then
         return;
      end if;
      Row := To_Item (Browser.Cache, Row);
      if Row = Null_Iter then
         return;
      end if;
      Name_Commit
      (  Browser,
         Get_Path (Browser.Cache, Row),
         Item_Name (Get_String (Nth (Params, 2)))
      );
   end Edited_Directory;

   procedure Edited_Item
             (  Object  : access GObject_Record'Class;
                Params : GValues;
                Column : Column_Data
             )  is
      Row : Gtk_Tree_Iter :=
               Get_Iter_From_String
               (  Column.Browser.Columns,
                  Get_String (Nth (Params, 1))
               );
   begin
      if Row = Null_Iter then
         return;
      end if;
      Row :=
         From_Columned
         (  Column.Browser.Columns,
            Row,
            Column.Column
         );
      if Row = Null_Iter then
         return;
      end if;
      Name_Commit
      (  Column.Browser,
         Positive (Get_Row_No (Column.Browser.Markup, Row) + 1),
         Item_Name (Get_String (Nth (Params, 2)))
      );
   end Edited_Item;

   procedure Emit
             (  Store  : access Gtk_Abstract_Directory_Record'Class;
                Signal : String;
                Text   : UTF8_String;
                Path   : Item_Path
             )  is
      procedure Internal
                (  Params : System.Address;
                   Signal : Signal_ID;
                   Quark  : GQuark;
                   Result : System.Address
                );
      pragma Import (C, Internal, "g_signal_emitv");
      Params : GValue_Array (0..2);
      Result : GValue;
   begin
      if Abstract_Directory_Class /= Uninitialized_Class then
         declare
            This : constant GType :=
                      Type_From_Class (Abstract_Directory_Class);
         begin
            Init (Params (0), This);
            Set_Object (Params (0), Store);
            Init (Params (1), GType_String);
            Set_String (Params (1), Text);
            Init (Params (2), GType_String);
            Set_String (Params (2), String (Path));
            Internal
            (  Params (0)'Address,
               Lookup (This, Signal),
               0,
               Result'Address
            );
            Unset (Params (0));
            Unset (Params (1));
            Unset (Params (2));
         end;
      end if;
   end Emit;

   procedure Emit
             (  Store  : access Gtk_Abstract_Directory_Record'Class;
                Signal : String;
                Text   : UTF8_String
             )  is
      procedure Internal
                (  Params : System.Address;
                   Signal : Signal_ID;
                   Quark  : GQuark;
                   Result : System.Address
                );
      pragma Import (C, Internal, "g_signal_emitv");
      Params : GValue_Array (0..1);
      Result : GValue;
   begin
      if Abstract_Directory_Class /= Uninitialized_Class then
         declare
            This : constant GType :=
                      Type_From_Class (Abstract_Directory_Class);
         begin
            Init (Params (0), This);
            Set_Object (Params (0), Store);
            Init (Params (1), GType_String);
            Set_String (Params (1), Text);
            Internal
            (  Params (0)'Address,
               Lookup (This, Signal),
               0,
               Result'Address
            );
            Unset (Params (0));
            Unset (Params (1));
         end;
      end if;
   end Emit;

   procedure Expand_Folder
             (  Store     : access Gtk_Abstract_Directory_Record'Class;
                Row       : Gtk_Tree_Iter;
                Directory : Item_Path
             )  is
      Updated : Boolean;
      Count   : GInt := 0;
      Item    : Gtk_Tree_Iter := Nth_Child (Store.Tree, Row, 0);
   begin
      while Item /= Null_Iter loop
         if Get_Int (Store.Tree, Item, 2) in Cached_Children then
            Add_Folder
            (  Store,
               Item,
               Get_Path
               (  Store,
                  Directory,
                  Item_Name (Get_String (Store.Tree, Item, 1))
               ),
               False,
               Updated
            );
            if Item /= Null_Iter then
               Count := Count + 1;
            end if;
         else
            Count := Count + 1;
         end if;
         Item := Nth_Child (Store.Tree, Row, Count);
      end loop;
   end Expand_Folder;

   function Filter
            (  Widget    : access Gtk_Directory_Items_View_Record;
               Directory : Boolean;
               Name      : Item_Name;
               Kind      : Item_Type
            )  return Boolean is
   begin
      return True;
   end Filter;

   function Filter_Leaves_Out
            (  Model : access Gtk_Tree_Model_Record'Class;
               Row   : Gtk_Tree_Iter;
               Store : Gtk_Abstract_Directory
            )  return Boolean is
   begin
      if Store.Inserting > 0 then
         return False;
      else
         return Get_Int (Store.Tree, Row, 2) in Cached_Directory;
      end if;
   end Filter_Leaves_Out;

   function Filter_List
            (  Model : access Gtk_Tree_Model_Record'Class;
               Row   : Gtk_Tree_Iter;
               Store : Gtk_Directory_Items_View
            )  return Boolean is
   begin
      if Store.Cache.Inserting > 0 or else Store.Columns = null then
         return False;
      elsif Row = Null_Iter then
         return True;
      elsif Get_Int (Store.Cache.Tree, Row, 2) in Cached_Directory then
         return True;
      end if;
      if Store.Root = null then
         if Parent (Store.Cache.Tree, Row) /= Null_Iter then
            return False;
         end if;
      else
         if (  Get_Iter (Store.Cache.Tree, Store.Root)
            /= Parent (Store.Cache.Tree, Row)
            )
         then
            return False;
         end if;
      end if;
      return
         Filter
         (  Store,
            False,
            Item_Name (Get_String  (Store.Cache.Tree, Row, 1)),
            Item_Type (Get_String  (Store.Cache.Tree, Row, 0))
         );
   end Filter_List;

   procedure Finalize
             (  Widget : access Gtk_Directory_Items_View_Record
             )  is
      procedure Free is
         new Ada.Unchecked_Deallocation
             (  Gtk_Cell_Renderer_Text_Array,
                Gtk_Cell_Renderer_Text_Array_Ptr
             );
   begin
      if Widget.Directories /= null then
         Unref (Widget.Directories);
      end if;
      if Widget.Root /= null then
         Path_Free (Widget.Root);
      end if;
      Free (Widget.Name_Renderers);
   end Finalize;

   procedure Finalize (Object : in out Selection_History) is
   begin
      if Object.Last /= null then
         Path_Free (Object.Last);
      end if;
      if Object.This /= null then
         Path_Free (Object.This);
      end if;
   end Finalize;

   function Find
            (  Store : access Gtk_Abstract_Directory_Record'Class;
               Path  : Item_Path
            )  return Gtk_Tree_Iter is
   begin
      if Path'Length = 0 then
         return Null_Iter;
      end if;
      declare
         Directory : Item_Path renames Get_Directory (Store, Path);
         Row       : Gtk_Tree_Iter := Find (Store, Directory);
      begin
         if Row /= Null_Iter then
            Row := Find (Store, Row, Directory, Get_Name (Store, Path));
         end if;
         return Row;
      end;
   exception
      when Name_Error =>
         return Find (Store, Null_Iter, "", Get_Name (Store, Path));
   end Find;

   function Find
            (  Store     : access Gtk_Abstract_Directory_Record'Class;
               Row       : Gtk_Tree_Iter;
               Directory : Item_Path;
               Name      : Item_Name
            )  return Gtk_Tree_Iter is
      Iter : Gtk_Tree_Iter;
      Aim  : Directory_Item :=
             (  Directory   => False,
                Policy      => Cache_Ahead,
                Name_Length => Name'Length,
                Name        => Name,
                Kind_Length => 0,
                Kind        => ""
             );
   begin
      if Row = Null_Iter then
         Iter := Get_Iter_First (Store.Tree);
      else
         Iter := Children (Store.Tree, Row);
      end if;
      while Iter /= Null_Iter loop
         declare
            Name : Item_Name :=
                      Item_Name (Get_String (Store.Tree, Iter, 1));
         begin
            exit when
                 (  Equal
                 =  Compare
                    (  Store,
                       Directory,
                       Aim,
                       (  Directory   => False,
                          Policy      => Cache_Ahead,
                          Name_Length => Name'Length,
                          Name        => Name,
                          Kind_Length => 0,
                          Kind        => ""
                       ),
                       True
                 )  );
         end;
         Next (Store.Tree, Iter);
      end loop;
      return Iter;
   end Find;

   function Find
            (  Store     : access Gtk_Abstract_Directory_Record'Class;
               Row       : Gtk_Tree_Iter;
               Directory : Item_Path;
               Item      : Directory_Item
            )  return Integer is
      Upper : GInt := N_Children (Store.Tree, Row);
      Lower : GInt := -1;
      This  : GInt;
      That  : Gtk_Tree_Iter;
   begin
      if Upper = 0 then
         return -1;
      end if;
      loop
         This := (Upper + Lower) / 2;
         That := Nth_Child (Store.Tree, Row, This);
         declare
            Cached : Directory_Item := Get_Item (Store, That);
         begin
            case Compare (Store, Directory, Item, Cached, False) is
               when Before =>
                  Upper := This;
                  if Upper - Lower <= 1 then
                     return -Integer (This + 1);
                  end if;
               when Equal =>
                  return Integer (This + 1);
               when After =>
                  Lower := This;
                  if Upper - Lower <= 1 then
                     return -Integer (This + 2);
                  end if;
            end case;
         end;
      end loop;
   end Find;

   function Find
            (  Store     : access Gtk_Abstract_Directory_Record'Class;
               Row       : Gtk_Tree_Iter;
               Directory : Item_Path;
               Item      : Directory_Item
            )  return Gtk_Tree_Iter is
      Position : constant Integer := Find (Store, Row, Directory, Item);
   begin
      if Position < 0 then
         return Null_Iter;
      else
         return Nth_Child (Store.Tree, Row, GInt (Position) - 1);
      end if;
   end Find;

   function From_Item
            (  Store : access Gtk_Abstract_Directory_Record'Class;
               Item  : Gtk_Tree_Iter
            )  return Gtk_Tree_Iter is
      Directory : Gtk_Tree_Iter;
   begin
      Convert_Child_Iter_To_Iter (Store, Directory, Item);
      return Directory;
   end From_Item;

   function From_Marked
            (  Widget : access Gtk_Directory_Items_View_Record;
               Item   : Gtk_Tree_Iter
            )  return Gtk_Tree_Iter is
      Row : Gtk_Tree_Iter;
   begin
      Convert_Iter_To_Child_Iter
      (  Widget.Content,
         Row,
         From_Extension (Widget.Markup, Item)
      );
      return Row;
   end From_Marked;

   function Get_Cache
            (  Widget : access Gtk_Directory_Tree_View_Record
            )  return Gtk_Abstract_Directory is
   begin
      return Widget.Cache;
   end Get_Cache;

   function Get_Cache (Widget : access Gtk_Directory_Items_View_Record)
      return Gtk_Abstract_Directory is
   begin
      return Widget.Cache;
   end Get_Cache;

   function Get_Cached
            (  Store : access Gtk_Abstract_Directory_Record;
               Path  : Item_Path
            )  return Directory_Item is
      Row : Gtk_Tree_Iter := Find (Store, Path);
   begin
      if Row = Null_Iter then
         raise Constraint_Error;
      else
         return Get_Item (Store, Row);
      end if;
   end Get_Cached;

   function Get_Column
            (  Widget : access Gtk_Directory_Items_View_Record;
               Index  : Positive
            )  return Positive is
      Item   : Gtk_Tree_Iter := Get_Iter (Widget, Index);
      Column : Positive;
   begin
      To_Columned (Widget.Columns, Item, Column);
      return Column;
   end Get_Column;

   function Get_Columns
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Positive is
   begin
      return Get_Major_Columns (Widget.Columns);
   end Get_Columns;

   function Get_Current
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Natural is
      Columned_Path : Gtk_Tree_Path;
      Markup_Path   : Gtk_Tree_Path;
      Column        : Gtk_Tree_View_Column;
      Column_No     : GInt;
   begin
      Get_Cursor (Widget, Columned_Path, Column);
      if Columned_Path = null then
         return 0;
      end if;
      Column_No := Get_Column_No (Widget, Column);
      if Column_No < 0 then
         Path_Free (Columned_Path);
         return 0;
      end if;
      Markup_Path :=
         From_Columned
         (  Widget.Columns,
            Columned_Path,
            Positive (Column_No + 1)
         );
      if Markup_Path = null then
         Path_Free (Columned_Path);
         return 0;
      end if;
      declare
         Indices : GInt_Array renames Get_Indices (Markup_Path);
         Result  : Natural := 0;
      begin
         if Indices'Length > 0 then
            Result := Natural (Indices (Indices'Last) + 1);
         end if;
         Path_Free (Markup_Path);
         Path_Free (Columned_Path);
         return Result;
      end;
   end Get_Current;

   function Get_Current_Directory
            (  Widget : access Gtk_Directory_Tree_View_Record
            )  return Item_Path is
      Row   : Gtk_Tree_Iter;
      Model : Gtk_Tree_Model;
   begin
      Get_Selected (Get_Selection (Widget), Model, Row);
      if Row = Null_Iter then
         raise Name_Error;
      end if;
      return Get_Path (Widget.Cache, To_Item (Widget.Cache, Row));
   end Get_Current_Directory;

   function Get_Directory
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Item_Path is
      Current : Gtk_Tree_Iter := Get_Directory (Widget);
   begin
      return Get_Path (Widget.Cache, Current);
   end Get_Directory;

   function Get_Directory
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Gtk_Tree_Iter is
   begin
      if Widget.Root = null then
         return Null_Iter;
      else
         return Get_Iter (Widget.Cache.Tree, Widget.Root);
      end if;
   end Get_Directory;

   function Get_Directory_Size
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Natural is
   begin
      return
         Natural
         (  N_Children
            (  Widget.Markup,
               Get_Root (Widget.Columns)
         )  );
   end Get_Directory_Size;

   function Get_Directory_Tree_View
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Gtk_Directory_Tree_View is
   begin
      return Widget.Directories;
   end Get_Directory_Tree_View;

   function Get_Icon
            (  Widget       : access Gtk_Directory_Tree_View_Record;
               Kind         : Item_Type;
               Expanded     : Boolean;
               Has_Children : Boolean
            )  return String is
      This : String := String (Kind);
   begin
      if Expanded and then This = Stock_Directory then
         return Stock_Open;
      else
         return This;
      end if;
   end Get_Icon;

   function Get_Icon
            (  Widget    : access Gtk_Directory_Items_View_Record;
               Name      : Item_Name;
               Kind      : Item_Type;
               Directory : Boolean
            )  return String is
   begin
      return String (Kind);
   end Get_Icon;

   function Get_Index
            (  Widget : access Gtk_Directory_Items_View_Record;
               Row    : Positive;
               Column : Positive
            )  return Natural is
   begin
      return
         Natural
         (  Get_Row_No
            (  Widget.Markup,
               Get_Reference_Iter (Widget.Columns, Row, Column)
            )
         +  1
         );
   end Get_Index;

   function Get_Index
            (  Widget : access Gtk_Directory_Items_View_Record;
               Name   : Item_Name
            )  return Natural is
      Row       : Gtk_Tree_Iter := Get_Directory (Widget);
      Directory : Item_Path renames Get_Directory (Widget);
      Index     : Natural := 0;
      Aim       : Directory_Item :=
                  (  Directory   => False,
                     Policy      => Cache_Ahead,
                     Name_Length => Name'Length,
                     Name        => Name,
                     Kind_Length => 0,
                     Kind        => ""
                  );
   begin
      Row := To_Marked (Widget, Row);
      if Row = Null_Iter then
         return 0;
      end if;
      Row := Children (Widget.Markup, Row);
      while Row /= Null_Iter loop
         Index := Index + 1;
         declare
            Name : Item_Name :=
                      Item_Name (Get_String (Widget.Markup, Row, 1));
         begin
            if (  Equal
               =  Compare
                  (  Widget.Cache,
                     Directory,
                     Aim,
                     (  Directory   => False,
                        Policy      => Cache_Ahead,
                        Name_Length => Name'Length,
                        Name        => Name,
                        Kind_Length => 0,
                        Kind        => ""
                     ),
                     True
               )  )
            then
               return Index;
            end if;
         end;
         Next (Widget.Markup, Row);
      end loop;
      return 0;
   end Get_Index;

   function Get_Item
            (  Store : access Gtk_Abstract_Directory_Record'Class;
               Item  : Gtk_Tree_Iter
            )  return Directory_Item is
      Kind : UTF8_String renames Get_String (Store.Tree, Item, 0);
      Name : UTF8_String renames Get_String (Store.Tree, Item, 1);
   begin
      return
      (  Policy      => Cache_Expanded,
         Name_Length => Name'Length,
         Name        => Item_Name (Name),
         Kind_Length => Kind'Length,
         Kind        => Item_Type (Kind),
         Directory   =>
            Get_Int (Store.Tree, Item, 2) in Cached_Directory
      );
   end Get_Item;

   function Get_Iter
            (  Widget : access Gtk_Directory_Items_View_Record;
               Index  : Positive
            )  return Gtk_Tree_Iter is
      Row : Gtk_Tree_Iter :=
               Nth_Child
               (  Widget.Markup,
                  Get_Root (Widget.Columns),
                  GInt (Index) - 1
               );
   begin
      if Row = Null_Iter then
         raise Constraint_Error;
      else
         return Row;
      end if;
   end Get_Iter;

   function Get_Name
            (  Widget : access Gtk_Directory_Items_View_Record;
               Index  : Positive
            )  return Item_Name is
   begin
      return
         Item_Name
         (  Get_String
            (  Widget.Markup,
               Get_Iter (Widget, Index),
               1
         )  );
   end Get_Name;

   function Get_Path
            (  Widget : access Gtk_Directory_Items_View_Record;
               Name   : Item_Name
            )  return Item_Path is
   begin
      return
         Get_Path
         (  Widget.Cache,
            Get_Directory (Widget),
            Name
         );
   end Get_Path;

   function Get_Path
            (  Store  : access Gtk_Abstract_Directory_Record'Class;
               Item   : Gtk_Tree_Iter
            )  return Item_Path is
   begin
      if Item = Null_Iter then
         return "";
      end if;
      declare
         Folder : Gtk_Tree_Iter := Parent (Store.Tree, Item);
         Name   : constant Item_Name :=
                     Item_Name (Get_String (Store.Tree, Item, 1));
      begin
         if Folder = Null_Iter then
            return Get_Path (Store, "", Name);
         else
            return Get_Path (Store, Get_Path (Store, Folder), Name);
         end if;
      end;
   end Get_Path;

   function Get_Path
            (  Widget : access Gtk_Directory_Items_View_Record;
               Index  : Positive
            )  return Item_Path is
   begin
      return
         Get_Path
         (  Widget.Cache,
            From_Marked (Widget, Get_Iter (Widget, Index))
         );
   end Get_Path;

   procedure Get_Position
             (  Widget : access Gtk_Directory_Items_View_Record;
                Index  : Positive;
                Row    : out Positive;
                Column : out Positive
             )  is
      No   : GInt;
      Iter : Gtk_Tree_Iter :=
                Nth_Child
                (  Widget.Markup,
                   Get_Root (Widget.Columns),
                   GInt (Index) - 1
                );
   begin
      if Iter /= Null_Iter then
         To_Columned (Widget.Columns, Iter, Column);
         No := Get_Row_No (Widget.Columns, Iter);
         if No >= 0 then
            Row := Positive (No + 1);
            return;
         end if;
      end if;
      raise Constraint_Error;
   end Get_Position;

   function Get_Row
            (  Widget : access Gtk_Directory_Items_View_Record;
               Index  : Positive
            )  return Positive is
      Item   : Gtk_Tree_Iter := Get_Iter (Widget, Index);
      Column : Positive;
   begin
      To_Columned (Widget.Columns, Item, Column);
      return Positive (Get_Row_No (Widget.Columns, Item) + 1);
   end Get_Row;

   function Get_Selection
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Selection is
      Result : Selection (1..Widget.Markup.Selected);
   begin
      if Result'Length > 0 then
         declare
            Root  : Gtk_Tree_Iter := Get_Root (Widget.Columns);
            Index : GInt := 0;
            Row   : Gtk_Tree_Iter;
         begin
            for Selected in Result'Range loop
               loop
                  Row := Nth_Child (Widget.Markup, Root, Index);
                  if Row = Null_Iter then
                     Log
                     (  GtkAda_Contributions_Domain,
                        Log_Level_Critical,
                        (  "Inconsitent selection in "
                        &  Where ("Get_Selection")
                     )  );
                     Widget.Markup.Selected := Selected - 1;
                     return Result (1..Widget.Markup.Selected);
                  end if;
                  Index := Index + 1;
                  exit when Get_Boolean (Widget.Markup, Row, 3);
               end loop;
               Result (Selected) := Positive (Index);
            end loop;
         end;
      end if;
      return Result;
   end Get_Selection;

   function Get_Selection_Mode
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Gtk_Selection_Mode is
   begin
      return Widget.Markup.Mode;
   end Get_Selection_Mode;

   function Get_Selection_Size
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Natural is
   begin
      return Widget.Markup.Selected;
   end Get_Selection_Size;

   function Get_Tree_Store
            (  Store : access Gtk_Abstract_Directory_Record
            )  return Gtk_Tree_Store is
   begin
      return Store.Tree;
   end Get_Tree_Store;

   function Get_Type
            (  Widget : access Gtk_Directory_Items_View_Record;
               Index  : Positive
            )  return Item_Type is
   begin
      return
         Item_Type
         (  Get_String
            (  Widget.Cache.Tree,
               From_Marked (Widget, Get_Iter (Widget, Index)),
               0
         )  );
   end Get_Type;

   function Get_Visible_Height
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Natural is
      Window : Gdk_Window := Get_Bin_Window (Widget);
   begin
      if Window = null then
         return 0;
      end if;
      declare
         Bottom : Natural;
         Top    : Natural;
         X_Size : GInt;
         Y_Size : GInt;
      begin
         Gdk.Drawable.Get_Size (Window, X_Size, Y_Size);
         Top := Locate (Widget, 0.5, 0.5);
         if Top = 0 then
            return 0;
         end if;
         Bottom := Locate (Widget, 0.5, GDouble (Y_Size) - 0.5);
         if Bottom = 0 then
            Bottom :=
               Get_Column_Height
               (  Widget.Columns,
                  Get_Column (Widget, Top)
               );
         end if;
         return Bottom - Top + 1;
      end;
   end Get_Visible_Height;

   function Get_Visible_Width
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Natural is
      Window : Gdk_Window := Get_Bin_Window (Widget);
   begin
      if Window = null then
         return 0;
      end if;
      declare
         Left   : Natural;
         Right  : Natural;
         X_Size : GInt;
         Y_Size : GInt;
      begin
         Gdk.Drawable.Get_Size (Window, X_Size, Y_Size);
         Left := Locate (Widget, 0.5, 0.5);
         if Left = 0 then
            return 0;
         end if;
         Right := Locate (Widget, GDouble (X_Size) - 0.5, 0.5);
         if Right = 0 then
            Right :=
               Get_Row_Width
               (  Widget.Columns,
                  Get_Row (Widget, Left)
               );
         end if;
         return Right - Left + 1;
      end;
   end Get_Visible_Width;

   procedure Gtk_New
             (  Widget  : out Gtk_Directory_Items_View;
                Store   : access Gtk_Abstract_Directory_Record'Class;
                Columns : Positive;
                Current : Item_Path := ""
             )  is
   begin
      Widget := new Gtk_Directory_Items_View_Record;
      Initialize (Widget, Store, Columns, Current);
   exception
      when others =>
         Destroy (Widget);
         raise;
   end Gtk_New;

   procedure Gtk_New
             (  Widget  : out Gtk_Directory_Items_View;
                Tree    : access Gtk_Directory_Tree_View_Record'Class;
                Columns : Positive
             )  is
   begin
      Widget := new Gtk_Directory_Items_View_Record;
      Initialize (Widget, Tree, Columns);
   exception
      when others =>
         Destroy (Widget);
         raise;
   end Gtk_New;

   procedure Gtk_New
             (  Widget   : out Gtk_Directory_Tree_View;
                Store    : access Gtk_Abstract_Directory_Record'Class;
                Selected : Item_Path := ""
             )  is
   begin
      Widget := new Gtk_Directory_Tree_View_Record;
      Initialize (Widget, Store, Selected);
   exception
      when others =>
         Destroy (Widget);
         raise;
   end Gtk_New;

   procedure Initialize
             (  Store : access Gtk_Abstract_Directory_Record'Class
             )  is
   begin
      Gtk_New
      (  Store.Tree,
         (  GType_String, -- Icon
            GType_String, -- Name
            GType_Int     -- Directory caching flag
      )  );
      -- Directories view
      Initialize (Store, Store.Tree);
      Initialize_Class_Record
      (  Store,
         Abstract_Directory_Signals,
         Abstract_Directory_Class,
         Abstract_Directory_Class_Name,
         (  0 => (0 => GType_String, 1 => GType_String),
            1 => (0 => GType_String, 1 => GType_String),
            2 => (0 => GType_String, 1 => GType_None)
      )  );
      -- Filling the root directory
      declare
         Updated : Boolean;
         Root    : Gtk_Tree_Iter := Null_Iter;
      begin
         Add_Folder (Store, Root, "", True, Updated);
      end;
      Directory_Filter.Set_Visible_Func
      (  Store,
         Filter_Leaves_Out'Access,
         Store.all'Access
      );
      Unref (Store.Tree);
   exception
      when others =>
         if Store.Tree /= null then
            Unref (Store.Tree);
         end if;
   end Initialize;

   procedure Initialize
             (  Widget   : access Gtk_Directory_Tree_View_Record'Class;
                Store    : access Gtk_Abstract_Directory_Record'Class;
                Selected : Item_Path
             )  is
   begin
      Widget.Cache := Store.all'Access;
      Gtk.Tree_View.Initialize (Widget);
      Set_Enable_Search (Widget, False);
      Set_Rules_Hint (Widget, True);
      Set_Headers_Visible (Widget, False);
      declare
         Column    : Gtk_Tree_View_Column;
         Column_No : GInt;
      begin
         Gtk_New (Column);

         Gtk_New (Widget.Icon_Renderer);
         Pack_Start (Column, Widget.Icon_Renderer, False);
         Add_Attribute (Column, Widget.Icon_Renderer, "stock-id", 0);

         Gtk_New (Widget.Name_Renderer);
         Pack_Start (Column, Widget.Name_Renderer, True);
         Add_Attribute (Column, Widget.Name_Renderer, "text", 1);
         Column_No := Append_Column (Widget, Column);
         Set_Resizable (Column, True);
         Tree_Functions.Set_Cell_Data_Func
         (  Column,
            Widget.Icon_Renderer,
            Set_Tree_Data'Access,
            Widget.all'Access
         );
      end;
      Set_Model (Widget, Store.all'Access);
      Set_Mode (Get_Selection (Widget), Selection_Single);
      Directory_Tree_Handlers.Connect
      (  Widget.Name_Renderer,
         "edited",
         Edited_Directory'Access,
         Widget.all'Access
      );
      Directory_Tree_Handlers.Connect
      (  Widget,
         "row-activated",
         Directory_Activated'Access,
         Widget.all'Access
      );
      Directory_Tree_Handlers.Connect
      (  Widget,
         "row-expanded",
         Directory_Expanded'Access,
         Widget.all'Access
      );
      Directory_Tree_Handlers.Connect
      (  Get_Selection (Widget),
         "changed",
         Directory_Changed'Access,
         Widget.all'Access
      );
      Directory_Tree_Handlers.Connect
      (  Widget.Cache,
         "refreshed",
         Directory_Refreshed'Access,
         Widget.all'Access
      );
      if Selected'Length > 0 then
         Set_Current_Directory (Widget, Selected);
      end if;
   end Initialize;

   procedure Initialize
             (  Widget  : access Gtk_Directory_Items_View_Record'Class;
                Store   : access Gtk_Abstract_Directory_Record'Class;
                Columns : Positive;
                Current : Gtk_Tree_Iter
             )  is
      Changed : Boolean := False;
   begin
      Gtk.Tree_View.Initialize (Widget);
      Initialize_Class_Record
      (  Widget,
         Directory_Items_Signals,
         Directory_Items_Class,
         Directory_Items_Class_Name
      );
      Set_Headers_Visible (Widget, False);
      Set_Enable_Search (Widget, False);
      Widget.Cache := Store.all'Access;
      Gtk_New (Widget.Content, Store.Tree);
      Widget.Markup := new Gtk_Selection_Store_Record;
      Widget.Markup.View := Widget.all'Access;
      Initialize (Widget.Markup, Widget.Content, (1 => GType_Boolean));
      Gtk_New
      (  Widget.Columns,
         Widget.Markup,
         Columns,
         To_Marked (Widget, Current)
      );
      Items_Filter.Set_Visible_Func
      (  Widget.Content,
         Filter_List'Access,
         Widget.all'Access
      );
      -- List view
      Widget.Name_Renderers :=
         new Gtk_Cell_Renderer_Text_Array (1..Columns);
      declare
         Column    : Gtk_Tree_View_Column;
         Column_No : GInt;
         Data      : Column_Data;
      begin
         Data.Browser := Widget.all'Access;
         Gtk_New (Data.Icon_Renderer);
         for No in 0..GInt (Columns) - 1 loop
            Gtk_New (Data.Text_Renderer);
            Widget.Name_Renderers (Integer (No + 1)) :=
               Data.Text_Renderer;
            Gtk_New (Column);

            Pack_Start (Column, Data.Icon_Renderer, False);
            Pack_Start (Column, Data.Text_Renderer, True);
            Column_No := Append_Column (Widget, Column);
--          Set_Sizing (Column, Tree_View_Column_Fixed);
            Set_Expand (Column, True);

            Data.Column := Positive (No + 1);
            Column_Functions.Set_Cell_Data_Func
            (  Column,
               Data.Icon_Renderer,
               Set_Column_Data'Access,
               Data
            );
            Directory_Items_Commit_Handlers.Connect
            (  Data.Text_Renderer,
               "edited",
               Edited_Item'Access,
               Data
            );
         end loop;
      end;
      Set_Mode (Get_Selection (Widget), Selection_None);
      if Widget.Directories /= null then
         Directory_Items_Result_Handlers.Connect
         (  Widget,
            "button_press_event",
            Directory_Items_Result_Handlers.To_Marshaller
               (Key_Press'Access),
            Widget.all'Access
         );
         Directory_Items_Handlers.Connect
         (  Widget,
            "destroy",
            Destroy'Access,
            Widget.all'Access
         );
         Directory_Selection_Handlers.Connect
         (  Get_Selection (Widget.Directories),
            "changed",
            Selection_Changed'Access,
            Widget.all'Access
         );
         Directory_Items_Result_Handlers.Connect
         (  Widget,
            "key_press_event",
            Directory_Items_Result_Handlers.To_Marshaller
               (Key_Press'Access),
            Widget.all'Access
         );
         Directory_Items_Handlers.Connect
         (  Widget,
            "row-activated",
            Item_Activated'Access,
            Widget.all'Access
         );
         Directory_Items_Handlers.Connect
         (  Widget.Directories,
            "row-collapsed",
            Directory_Collapsed'Access,
            Widget.all'Access
         );
      end if;
      Set_Model (Widget, Widget.Columns.all'Access);
      Set_Current (Widget, Current, Changed);
      Unref (Widget.Columns);
      Unref (Widget.Content);
      Unref (Widget.Markup);
      if Changed  then
         Directory_Changed (Widget);
      end if;
   exception
      when others =>
         Unref (Widget.Columns);
         Unref (Widget.Content);
         Unref (Widget.Markup);
         if Widget.Directories /= null then
            Unref (Widget.Directories);
         end if;
         raise;
   end Initialize;

   procedure Initialize
             (  Widget  : access Gtk_Directory_Items_View_Record'Class;
                Store   : access Gtk_Abstract_Directory_Record'Class;
                Columns : Positive;
                Current : Item_Path
             )  is
      Best_Match, Exact_Match : Gtk_Tree_Iter;
   begin
      Cache (Store, Current, Best_Match, Exact_Match);
      Initialize (Widget, Store, Columns, Best_Match);
   end Initialize;

   procedure Initialize
             (  Widget  : access Gtk_Directory_Items_View_Record'Class;
                Tree    : access Gtk_Directory_Tree_View_Record'Class;
                Columns : Positive
             )  is
      Row   : Gtk_Tree_Iter;
      Model : Gtk_Tree_Model;
   begin
      Get_Selected (Get_Selection (Tree), Model, Row);
      Row := To_Item (Tree.Cache, Row);
      Widget.Directories := Tree.all'Access;
      Ref (Widget.Directories);
      Initialize (Widget, Tree.Cache, Columns, Row);
   end Initialize;

   function Input_Event
            (  Widget : access Gtk_Directory_Items_View_Record;
               Index  : Positive;
               Event  : Gdk_Event_Key
            )  return Boolean is
   begin
      return False;
   end Input_Event;

   procedure Inserted
             (  Model : access Gtk_Selection_Store_Record;
                Path  : Gtk_Tree_Path;
                Iter  : Gtk_Tree_Iter
             )  is
   begin
      Inserted
      (  Gtk_Extension_Store_Record (Model.all)'Access,
         Path,
         Iter
      );
      if (  Model.Mode = Selection_Browse
         and then
            Model.Selected = 0
         )
      then
         -- Select one item
         declare
            Current : Natural := Get_Current (Model.View);
         begin
            if Current > 0 then
               Change_Selection (Model.View, Current, True);
            else
               Change_Selection (Model.View, 1, True);
            end if;
         end;
         Selection_Changed (Model.View);
      end if;
   end Inserted;

   function Is_Directory
            (  Widget : access Gtk_Directory_Items_View_Record;
               Index  : Positive
            )  return Boolean is
   begin
      return
      (  Get_Int
         (  Widget.Cache.Tree,
            From_Marked (Widget, Get_Iter (Widget, Index)),
            2
         )
      in Cached_Directory
      );
   end Is_Directory;

   function Is_Editable
            (  Widget : access Gtk_Directory_Items_View_Record
            )  return Boolean is
   begin
      return
         Get_Property
         (  Widget.Name_Renderers (1),
            Editable_Property
         );
   end Is_Editable;

   function Is_Editable
            (  Widget : access Gtk_Directory_Tree_View_Record
            )  return Boolean is
   begin
      return Get_Property (Widget.Name_Renderer, Editable_Property);
   end Is_Editable;

   function Is_Selected
            (  Widget : access Gtk_Directory_Items_View_Record;
               Index  : Positive
            )  return Boolean is
   begin
      return
         Get_Boolean (Widget.Markup, Get_Iter (Widget, Index), 3);
   end Is_Selected;

   procedure Item_Activated
             (  Object  : access GObject_Record'Class;
                Params  : GValues;
                Browser : Gtk_Directory_Items_View
             )  is
      Path   : Gtk_Tree_Path := Convert (Get_Address (Nth (Params, 1)));
      Column : Gtk_Tree_View_Column := Get_Column (Nth (Params, 2));
   begin
      if Path = null or else Column = null then
         return;
      else
         Activated
         (  Browser,
            Get_Index
            (  Browser,
               Positive (Get_Row_No (Browser.Columns, Path) + 1),
               Positive (Get_Column_No (Browser, Column) + 1)
         )  );
      end if;
   end Item_Activated;

   function Key_Press
            (  Widget  : access Gtk_Tree_View_Record'Class;
               Event   : Gdk_Event_Key;
               Browser : Gtk_Directory_Items_View
            )  return Boolean is
      Changed : Boolean;
   begin
      case Get_Event_Type (Event) is
         when Button_Press =>
            -- Button click
            declare
               Index : constant Natural :=
                       Locate (Browser, Get_X (Event), Get_Y (Event));
            begin
               if Get_Button (Event) = 1 then
                  Move
                  (  Browser,
                     Changed,
                     Get_State (Event),
                     Index
                  );
                  return not Changed;
               else
                  if Index > 0 then
                     return Input_Event (Browser, Index, Event);
                  else
                     return False;
                  end if;
               end if;
            end;
         when Gdk_2Button_Press | Gdk_3Button_Press =>
            -- Double click
            declare
               Index : constant Natural :=
                  Locate (Browser, Get_X (Event), Get_Y (Event));
            begin
               if Index > 0 then
                  Activated (Browser, Index);
               end if;
               return True;
            end;
         when Key_Press =>
            -- Key press
            case Get_Key_Val (Event) is
               when GDK_Up | GDK_KP_Up =>
                  Move
                  (  Browser,
                     Changed,
                     Get_State (Event),
                     Get_Current (Browser),
                     -1
                  );
                  return True;
               when GDK_Down | GDK_KP_Down =>
                  Move
                  (  Browser,
                     Changed,
                     Get_State (Event),
                     Get_Current (Browser),
                     1
                  );
                  return True;
               when GDK_Left | GDK_KP_Left =>
                  Move
                  (  Browser,
                     Changed,
                     Get_State (Event),
                     Get_Current (Browser),
                     -Get_Rows (Browser.Columns, False),
                     True
                  );
                  return True;
               when GDK_Page_Down | GDK_KP_Page_Down =>
                  Move
                  (  Browser,
                     Changed,
                     Get_State (Event),
                     Get_Current (Browser),
                     Get_Visible_Height (Browser)
                  );
                  return True;
               when GDK_Page_Up | GDK_KP_Page_Up =>
                  Move
                  (  Browser,
                     Changed,
                     Get_State (Event),
                     Get_Current (Browser),
                     -Get_Visible_Height (Browser)
                  );
                  return True;
               when GDK_Right | GDK_KP_Right =>
                  Move
                  (  Browser,
                     Changed,
                     Get_State (Event),
                     Get_Current (Browser),
                     Get_Rows (Browser.Columns, False),
                     True
                  );
                  return True;
               when GDK_Home | GDK_KP_Home =>
                  Move (Browser, Changed, Get_State (Event), 1);
                  return True;
               when GDK_End | GDK_KP_End =>
                  Move
                  (  Browser,
                     Changed,
                     Get_State (Event),
                     Get_Directory_Size (Browser)
                  );
                  return True;
               when others =>
                  -- Not a special key
                  declare
                     Key : GUnichar :=
                              Keyval_To_Unicode (Get_Key_Val (Event));
                  begin
                     if Key = 0 then
                        return False;
                     else
                        if Browser.Last_Key /= Key then
                           Browser.Last_Key       := Key;
                           Browser.Last_Position  := 0;
                        end if;
                        declare
                           Prefix : UTF8_String (1..8);
                           Last   : Natural;
                        begin
                           Unichar_To_UTF8 (Key, Prefix, Last);
                           Browser.Last_Position :=
                              Scan
                              (  Browser,
                                 Item_Name (Prefix (1..Last)),
                                 Browser.Last_Position
                              );
                        end;
                        if Browser.Last_Position /= 0 then
                           Move
                           (  Browser,
                              Changed,
                              Get_State (Event),
                              Browser.Last_Position
                           );
                           return True;
                        end if;
                     end if;
                  end;
            end case;
         when others =>
            null;
      end case;
      declare
--         Screen   : Gdk.Screen.Gdk_Screen;
--         Screen_X : GInt;
--         Screen_Y : GInt;
--         Window_X : GInt;
--         Window_Y : GInt;
--         Mask     : Gdk_Modifier_Type;
         Index    : Natural;
      begin
--         Get_Pointer (Get_Default, Screen, Screen_X, Screen_Y, Mask);
--         Gdk.Window.Get_Position
--         (  Get_Window (Widget),
--            Window_X,
--            Window_Y
--         );
--         Index :=
--            Locate
--            (  Browser,
--               GDouble (Screen_X - Window_X),
--               GDouble (Screen_Y - Window_Y)
--            );
         Index := Locate (Browser, Get_X (Event), Get_Y (Event));
         if Index > 0 then
            return Input_Event (Browser, Index, Event);
         else
            return False;
         end if;
      end;
   end Key_Press;

   function Locate
            (  Widget : access Gtk_Directory_Items_View_Record;
               X, Y   : GDouble
            )  return Natural is
      Cell_X        : GInt;
      Cell_Y        : GInt;
      Column        : Gtk_Tree_View_Column;
      Columned_Path : Gtk_Tree_Path;
      Column_No     : GInt;
      Markup_Path   : Gtk_Tree_Path;
      Found         : Boolean;
      Result        : Natural := 0;
   begin
      begin
         Get_Path_At_Pos
         (  Widget,
            GInt (X),
            GInt (Y),
            Columned_Path,
            Column,
            Cell_X,
            Cell_Y,
            Found
         );
      exception
         when Constraint_Error =>
            return 0;
      end;
      if Found then
         Column_No := Get_Column_No (Widget, Column);
         if Column_No >= 0 then
            Markup_Path :=
               From_Columned
               (  Widget.Columns,
                  Columned_Path,
                  Positive (Column_No + 1)
               );
            if Markup_Path /= null then
               declare
                  Indices : GInt_Array renames
                            Get_Indices (Markup_Path);
               begin
                  if Indices'Length > 0 then
                     Result := Natural (Indices (Indices'Last) + 1);
                  end if;
               end;
               Path_Free (Markup_Path);
            end if;
         end if;
         Path_Free (Columned_Path);
      end if;
      return Result;
   end Locate;

   procedure Move
             (  Widget   : access Gtk_Directory_Items_View_Record'Class;
                Changed      : out Boolean;
                Modifier     : Gdk_Modifier_Type;
                To           : Natural;
                By           : Integer := 0;
                Fixed_Row    : Boolean := False;
                Fixed_Column : Boolean := False
             )  is
      Size     : constant Natural := Get_Directory_Size (Widget);
      Position : Integer;
   begin
      Changed := False;
      -- Determining the target position
      if To not in 1..Size then
         return;
      end if;
      if Fixed_Row or else Fixed_Column then
         declare
            New_Row    : Positive;
            New_Column : Positive;
            Old_Row    : Positive;
            Old_Column : Positive;
         begin
            Get_Position (Widget, To, Old_Row, Old_Column);
            Position := To + By;
            if Position > Get_Directory_Size (Widget) then
               Position := Get_Directory_Size (Widget);
            elsif Position < 1 then
               Position := 1;
            end if;
            Get_Position (Widget, Position, New_Row, New_Column);
            if Fixed_Row and then New_Row /= Old_Row then
               -- The row differs
               if By > 0 then
                  -- Moving  forwards using the last available column to
                  -- in the row.
                  New_Column := Get_Row_Width (Widget.Columns, New_Row);
               else
                  -- Moving backwards using the first column
                  New_Column := 1;
               end if;
               if Fixed_Column and then New_Column /= Old_Column then
                  return;
               end if;
               Position := Get_Index (Widget, Old_Row, New_Column);
            elsif Fixed_Column and then New_Column /= Old_Column then
               if By > 0 then
                  -- Moving forwards using the last available row
                  New_Row :=
                     Get_Column_Height (Widget.Columns, New_Column);
               else
                  -- Moving backwards using the first row
                  New_Row := 1;
               end if;
               if Fixed_Row and then New_Row /= Old_Row then
                  return;
               end if;
               Position := Get_Index (Widget, New_Row, Old_Column);
            end if;
         end;
      else
         Position := To + By;
         if Position > Get_Directory_Size (Widget) then
            Position := Get_Directory_Size (Widget);
         elsif Position < 1 then
            Position := 1;
         end if;
      end if;
      -- Dealing with selection
      case Widget.Markup.Mode is
         when Selection_None =>
            null;
         when Selection_Single =>
            if 0 = (Modifier and Control_Mask) then
               -- Simple selection
               if not Is_Selected (Widget, Position) then
                  if Get_Selection_Size (Widget) > 0 then
                     -- Deselect any selected
                     for Index in 1..Size loop
                        if Is_Selected (Widget, Index) then
                           Change_Selection (Widget, Index, False);
                           exit;
                        end if;
                     end loop;
                  end if;
                  Change_Selection (Widget, Position, True);
                  Changed := True;
               end if;
            else
               -- Selection toggling
               if Is_Selected (Widget, Position) then
                  Change_Selection (Widget, Position, False);
               else
                  if Get_Selection_Size (Widget) > 0 then
                     -- Deselect any selected
                     for Index in 1..Size loop
                        if Is_Selected (Widget, Index) then
                           Change_Selection (Widget, Index, False);
                           exit;
                        end if;
                     end loop;
                  end if;
                  Change_Selection (Widget, Position, True);
               end if;
               Changed := True;
            end if;
         when Selection_Browse =>
            if not Is_Selected (Widget, Position) then
               -- Deselect any selected
               for Index in 1..Size loop
                  if Is_Selected (Widget, Index) then
                     Change_Selection (Widget, Index, False);
                     exit;
                  end if;
               end loop;
               Change_Selection (Widget, Position, True);
               Changed := True;
            end if;
         when Selection_Multiple =>
            if 0 = (Modifier and Shift_Mask) then
               if 0 = (Modifier and Control_Mask) then
                  -- Single selection
                  for Index in 1..Size loop
                     if Is_Selected (Widget, Index) then
                        if Index /= Position then
                           Change_Selection (Widget, Index, False);
                           Changed := True;
                        end if;
                     else
                        if Index = Position then
                           Change_Selection (Widget, Index, True);
                           Changed := True;
                        end if;
                     end if;
                  end loop;
               else
                  -- Toggling selection
                  if Is_Selected (Widget, Position) then
                     Change_Selection (Widget, Position, False);
                  else
                     Change_Selection (Widget, Position, True);
                  end if;
                  Changed := True;
               end if;
            else
               -- Range selection in nearest to Position
               declare
                  Above : Integer := 0;
                  Below : Integer := 0;
                  From  : Integer := Position;
                  To    : Integer := Position;
               begin
                  if not Is_Selected (Widget, Position) then
                     Above := Integer'Last;
                     while From > 1 loop
                        From := From - 1;
                        if Is_Selected (Widget, From) then
                           Above := Position - From;
                           exit;
                        end if;
                     end loop;
                     Below := Integer'Last;
                     while To < Size loop
                        To := To + 1;
                        if Is_Selected (Widget, To) then
                           Below := Below - Position;
                           exit;
                        end if;
                     end loop;
                  end if;
                  while From > 1 loop
                     exit when not Is_Selected (Widget, From - 1);
                     From := From - 1;
                  end loop;
                  while To < Size loop
                     exit when not Is_Selected (Widget, To + 1);
                     To := To + 1;
                  end loop;
                  if Above < Below then
                     To := Position;
                  elsif Above > Below then
                     From := Position;
                  elsif Above = 0 and then By /= 0 then
                     if By > 0 then
                        From := Position;
                     else
                        To := Position;
                     end if;
                  else
                     if Position - From > To - Position then
                        To := Position;
                     else
                        From := Position;
                     end if;
                  end if;
                  if 0 = (Modifier and Control_Mask) then
                     -- Deselect anything before From
                     for Index in 1..From - 1 loop
                        if Is_Selected (Widget, Index) then
                           Change_Selection (Widget, Index, False);
                           Changed := True;
                        end if;
                     end loop;
                     -- Deselect anything after To
                     for Index in To + 1..Size loop
                        if Is_Selected (Widget, Index) then
                           Change_Selection (Widget, Index, False);
                           Changed := True;
                        end if;
                     end loop;
                  end if;
                  -- Select everything in From..To
                  for Index in From..To loop
                     if not Is_Selected (Widget, Index) then
                        Change_Selection (Widget, Index, True);
                        Changed := True;
                     end if;
                  end loop;
               end;
            end if;
      end case;
      -- Moving to the target position
      if Get_Current (Widget) /= Position then
         -- Set position to the target
         declare
            Column_No : Positive;
            Column    : Gtk_Tree_View_Column;
            Iter      : Gtk_Tree_Iter := Get_Iter (Widget, Position);
            Path      : Gtk_Tree_Path;
         begin
            To_Columned (Widget.Columns, Iter, Column_No);
            Path := Get_Path (Widget.Columns, Iter);
            Column := Get_Column (Widget, GInt (Column_No) - 1);
            Set_Cursor_On_Cell
            (  Tree_View     => Widget,
               Path          => Path,
               Focus_Column  => Column,
               Start_Editing => False,
               Focus_Cell    =>
                  Widget.Name_Renderers (Column_No).all'Access
            );
            Scroll_To_Cell (Widget, Path, Column, False, 0.0, 0.0);
            Grab_Focus (Widget);
            Path_Free (Path);
         end;
      end if;
      if Changed then
         Selection_Changed (Widget);
      end if;
   end Move;

   procedure Name_Commit
             (  Widget : access Gtk_Directory_Tree_View_Record;
                Old_Path : Item_Path;
                New_Name : Item_Name
             )  is
   begin
      null;
   end Name_Commit;

   procedure Name_Commit
             (  Widget   : access Gtk_Directory_Items_View_Record;
                Index  : Positive;
                Name   : Item_Name
             )  is
   begin
      null;
   end Name_Commit;

   procedure Read_Error
             (  Store : access Gtk_Abstract_Directory_Record'Class;
                Text  : UTF8_String;
                Path  : Item_Path
             )  is
   begin
      Emit (Store, "read-error", Text, Path);
   end Read_Error;

   procedure Rewind_Error
             (  Store : access Gtk_Abstract_Directory_Record'Class;
                Text  : UTF8_String;
                Path  : Item_Path
             )  is
   begin
      Emit (Store, "rewind-error", Text, Path);
   end Rewind_Error;

   procedure Refilter
             (  Widget : access Gtk_Directory_Items_View_Record'Class
             )  is
   begin
      Refilter (Widget.Content);
   end Refilter;

   procedure Refresh
             (  Store : access Gtk_Abstract_Directory_Record'Class;
                Item  : in out Gtk_Tree_Iter
             )  is
      procedure Internal (Object : System.Address; Name : String);
      pragma Import (C, Internal, "ada_g_signal_emit_by_name");
      Updated : Boolean;
      Child   : Gtk_Tree_Iter;
   begin
      loop
         if Item = Null_Iter then
            Child := Get_Iter_First (Store.Tree);
         else
            Child := Children (Store.Tree, Item);
         end if;
         exit when Child = Null_Iter;
         Remove (Store.Tree, Child);
      end loop;
      if Item = Null_Iter then
         Add_Folder (Store, Item, "", True, Updated);
         Store.Refreshing := Store.Refreshing + 1;
         Emit (Store, "refreshed", "");
         Store.Refreshing := Store.Refreshing - 1;
      else
         declare
            Path : constant Item_Path := Get_Path (Store, Item);
         begin
            Add_Folder (Store, Item, Path, True, Updated);
            Store.Refreshing := Store.Refreshing + 1;
            Emit (Store, "refreshed", String (Path));
            Store.Refreshing := Store.Refreshing - 1;
         end;
      end if;
   end Refresh;

   procedure Renamed
             (  Store    : access Gtk_Abstract_Directory_Record'Class;
                Old_Path : Item_Path;
                New_Name : Item_Name
             )  is
      Row : Gtk_Tree_Iter := Find (Store, Old_Path);
   begin
      if Row = Null_Iter then
         raise Constraint_Error;
      end if;
      Set (Store.Tree, Row, 1, String (New_Name));
   end Renamed;

   procedure Renamed
             (  Widget : access Gtk_Directory_Items_View_Record'Class;
                Index  : in out Natural;
                Name   : Item_Name
             )  is
      Row : Gtk_Tree_Iter := Get_Iter (Widget, Index);
   begin
      Row := From_Marked (Widget, Row);
      Set (Widget.Cache.Tree, Row, 1, String (Name));
      Row := To_Marked (Widget, Row);
      if Row = Null_Iter then
         Index := 0;
      else
         Index := Positive (Get_Row_No (Widget.Markup, Row) + 1);
      end if;
   end Renamed;

   procedure Reset_Selection
             (  Widget : access Gtk_Directory_Items_View_Record'Class
             )  is
      Changed : Boolean := False;
   begin
      case Widget.Markup.Mode is
         when Selection_None | Selection_Browse =>
            null;
         when Selection_Single =>
            for Index in 1..Get_Directory_Size (Widget) loop
               if Is_Selected (Widget, Index) then
                  Change_Selection (Widget, Index, False);
                  Changed := True;
                  exit;
               end if;
            end loop;
         when Selection_Multiple =>
            for Index in 1..Get_Directory_Size (Widget) loop
               if Is_Selected (Widget, Index) then
                  Change_Selection (Widget, Index, False);
                  Changed := True;
               end if;
            end loop;
      end case;
      if Changed then
         Selection_Changed (Widget);
      end if;
   end Reset_Selection;

   function Scan
            (  Widget : access Gtk_Directory_Items_View_Record'Class;
               Prefix : Item_Name;
               Index  : Natural
            )  return Natural is
      Row       : constant Gtk_Tree_Iter :=
                     To_Marked  (Widget, Get_Directory (Widget));
      Directory : Item_Path renames Get_Directory (Widget);
      Aim       : Directory_Item :=
                  (  Directory   => False,
                     Policy      => Cache_Ahead,
                     Kind_Length => 0,
                     Kind        => "",
                     Name_Length => Prefix'Length,
                     Name        => Prefix
                  );
      function Is_Prefix (Iter : Gtk_Tree_Iter) return Boolean is
         Name : constant Item_Name :=
                   Item_Name (Get_String (Widget.Markup, Iter, 1));
      begin
         if Name'Length < Aim.Name_Length then
            return False;
         else
            return
            (  Equal
            =  Compare
               (  Widget.Cache,
                  Directory,
                  Aim,
                  (  Directory   => False,
                     Policy      => Cache_Ahead,
                     Kind_Length => 0,
                     Kind        => "",
                     Name_Length => Aim.Name_Length,
                     Name        => Name (1..Aim.Name_Length)
                  ),
                  True
            )  );
         end if;
      end Is_Prefix;

      Iter : Gtk_Tree_Iter :=
                Nth_Child (Widget.Markup, Row, GInt (Index));
      Position : Natural := Index;
   begin
      while Iter /= Null_Iter loop
         Position := Position + 1;
         if Is_Prefix (Iter) then
            return Position;
         end if;
         Next (Widget.Markup, Iter);
      end loop;
      return 0;
   end Scan;

   procedure Selection_Changed
             (  Selection : access Gtk_Tree_Selection_Record'Class;
                Browser   : Gtk_Directory_Items_View
             )  is
   begin
      if Browser.Cache.Inserting = 0 and then not Browser.Selecting then
         Browser.Selecting := True;
         declare
            Model     : Gtk_Tree_Model;
            Directory : Gtk_Tree_Iter;
            Changed   : Boolean := False;
         begin
            Get_Selected (Selection, Model, Directory);
            if Directory = Null_Iter then
               -- Try to cache the root directory we are switching to
               Add_Folder
               (  Browser.Cache,
                  Directory,
                  "",
                  False,
                  Changed
               );
               Set_Current (Browser, Null_Iter, Changed);
            else
               Directory := To_Item (Browser.Cache, Directory);
               if Get_Int (Browser.Cache.Tree, Directory, 2)
                  not in Cached_Children
               then
                  -- Try to cache the directory we are switching to
                  Add_Folder
                  (  Browser.Cache,
                     Directory,
                     Get_Path (Browser.Cache, Directory),
                     False,
                     Changed
                  );
               end if;
               Set_Current (Browser, Directory, Changed);
               Select_Iter
               (  Selection,
                  From_Item (Browser.Cache, Get_Directory (Browser))
               );
            end if;
            if Changed then
               Directory_Changed (Browser);
            end if;
         exception
            when Error : others =>
               Log
               (  GtkAda_Contributions_Domain,
                  Log_Level_Critical,
                  (  "Fault: "
                  &  Exception_Information (Error)
                  &  Where ("Selection_Changed")
               )  );
         end;
         Browser.Selecting := False;
      end if;
   end Selection_Changed;

   procedure Selection_Changed
             (  Widget : access Gtk_Directory_Items_View_Record
             )  is
   begin
      Directory_Tree_Handlers.Emit_By_Name
      (  Widget,
         "selection-changed"
      );
   end Selection_Changed;

   procedure Set_Column_Data
             (  Column : access Gtk_Tree_View_Column_Record'Class;
                Cell   : access Gtk_Cell_Renderer_Record'Class;
                Model  : access Gtk_Tree_Model_Record'Class;
                Iter   : Gtk_Tree_Iter;
                Data   : Column_Data
             )  is
      FG, BG : Gdk_Color;
      Style  : Gtk_Style;
      Offset : GInt   := GInt (Data.Column - 1) * 4;
      Name   : String := Get_String (Model, Iter, Offset + 1);
   begin
      Set_Property (Data.Text_Renderer, BG_Set,      True);
      Set_Property (Data.Text_Renderer, Cell_BG_Set, True);
      Set_Property (Data.Text_Renderer, FG_Set,      True);
      Style := Get_Style (Data.Browser);
      if Get_Boolean (Model, Iter, Offset + 3) then
         if Has_Focus_Is_Set (Data.Browser) then
            FG := Get_Text (Style, State_Selected);
            BG := Get_Base (Style, State_Selected);
         else
            FG := Get_Text (Style, State_Active);
            BG := Get_Base (Style, State_Active);
         end if;
      else
         FG := Get_Text (Style, State_Normal);
         BG := Get_Base (Style, State_Normal);
      end if;
      Glib.Properties.Set_Property
      (  Data.Text_Renderer,
         FG_GDK,
         FG'Address
      );
      Glib.Properties.Set_Property
      (  Data.Text_Renderer,
         BG_GDK,
         BG'Address
      );
      Glib.Properties.Set_Property
      (  Data.Text_Renderer,
         Cell_BG_GDK,
         BG'Address
      );
      Set_Property
      (  Data.Icon_Renderer,
         Stock_ID_Property,
         Get_Icon
         (  Data.Browser,
            Item_Name (Name),
            Item_Type (Get_String (Model, Iter, Offset)),
            Get_Int (Model, Iter, Offset + 2) in Cached_Directory
      )  );
      Set_Property
      (  Data.Text_Renderer,
         Text_Property,
         Name
      );
   end Set_Column_Data;

   procedure Set_Current
             (  Widget : access Gtk_Directory_Items_View_Record'Class;
                Directory : Gtk_Tree_Iter;
                Changed   : in out Boolean
             )  is
   begin
      Changed := Changed or Directory /= Get_Directory (Widget);
      if Changed or else Widget.Cache.Refreshing > 0 then
         declare
            Columns : constant Positive :=
                               Get_Major_Columns (Widget.Columns);
         begin
            Ref (Widget.Markup);
            Ref (Widget.Content);
            Set_Null_Reference (Widget.Columns);
            Set_Null_Reference (Widget.Markup);
            if Widget.Markup.Selected /= 0 then
               Log
               (  GtkAda_Contributions_Domain,
                  Log_Level_Critical,
                  (  "Persisting selection removed"
                  &  Where ("Set_Current")
               )  );
               Widget.Markup.Selected := 0;
            end if;
            if Widget.Root /= null then
               Path_Free (Widget.Root);
               Widget.Root := null;
            end if;
            if Directory /= Null_Iter then
               Widget.Root := Get_Path (Widget.Cache.Tree, Directory);
            end if;
            Refilter (Widget);
            Set_Reference (Widget.Markup, Widget.Content);
            Set_Reference
            (  Widget.Columns,
               Widget.Markup,
               Columns,
               To_Marked (Widget, Directory)
            );
            Unref (Widget.Content);
            Unref (Widget.Markup);
            Columns_Autosize (Widget);
         end;
      end if;
   end Set_Current;

   procedure Set_Current_Directory
             (  Widget    : access Gtk_Directory_Tree_View_Record;
                Directory : Item_Path
             )  is
      Best_Match  : Gtk_Tree_Iter;
      Exact_Match : Gtk_Tree_Iter;
   begin
      Cache (Widget.Cache, Directory, Best_Match, Exact_Match);
      if Best_Match /= Null_Iter then
         Exact_Match := From_Item (Widget.Cache, Best_Match);
         while Exact_Match /= Null_Iter loop
            exit when
                 (  Get_Int (Widget.Cache, Exact_Match, 2)
                 in Cached_Children
                 );
            Exact_Match := Parent (Widget.Cache, Exact_Match);
         end loop;
         if Exact_Match /= Null_Iter then
            Best_Match := To_Item (Widget.Cache, Exact_Match);
            declare
               Path : Gtk_Tree_Path;
            begin
               Path := Get_Path (Widget.Cache, Exact_Match);
               Expand_To_Path (Widget, Path);
               Path_Free (Path);
               Exact_Match := From_Item (Widget.Cache, Best_Match);
                  -- Has to be done twice, who knows why...
               Path := Get_Path (Widget.Cache, Exact_Match);
               Expand_To_Path (Widget, Path);
               Path_Free (Path);
               Exact_Match := From_Item (Widget.Cache, Best_Match);

               Path := Get_Path (Widget.Cache, Exact_Match);
               Scroll_To_Cell
               (  Widget,
                  Path,
                  Get_Column (Widget, 0),
                  False,
                  0.0,
                  0.0
               );
               Path_Free (Path);
               Select_Iter (Get_Selection (Widget), Exact_Match);
            end;
            return;
         end if;
      end if;
      Unselect_All (Get_Selection (Widget));
   end Set_Current_Directory;

   procedure Set_Editable
             (  Widget   : access Gtk_Directory_Tree_View_Record;
                Editable : Boolean
             )  is
   begin
      Set_Property
      (  Widget.Name_Renderer,
         Editable_Property,
         Editable
      );
   end Set_Editable;

   procedure Set_Editable
             (  Widget   : access Gtk_Directory_Items_View_Record;
                Editable : Boolean
             )  is
   begin
      for Index in Widget.Name_Renderers'Range loop
         Set_Property
         (  Widget.Name_Renderers (Index),
            Editable_Property,
            Editable
         );
      end loop;
   end Set_Editable;

   procedure Set_Item
             (  Store : access Gtk_Abstract_Directory_Record'Class;
                Row   : Gtk_Tree_Iter;
                Item  : Directory_Item
             )  is
      Flag : GInt;
   begin
      Store.Inserting := Store.Inserting + 1;
      Set (Store.Tree, Row, 1, UTF8_String (Item.Name));
      Set (Store.Tree, Row, 0, UTF8_String (Item.Kind));
      Store.Inserting := Store.Inserting - 1;
      if Item.Directory then
         case Item.Policy is
            when Cache_Never =>
               Flag := Cached_Never_Directory;
            when Cache_Expanded =>
               Flag := Cached_Expanded_Directory;
            when Cache_Ahead =>
               Flag := Cached_Ahead_Directory;
         end case;
      else
         Flag := Cached_Item;
      end if;
      Set (Store.Tree, Row, 2, Flag);
   end Set_Item;

   procedure Set_Selection
             (  Widget : access Gtk_Directory_Items_View_Record'Class;
                Index  : Positive;
                State  : Boolean
             )  is
      Size    : constant Natural := Get_Directory_Size (Widget);
      Changed : Boolean := False;
   begin
      if Index > Size then
         return;
      end if;
      case Widget.Markup.Mode is
         when Selection_None =>
            null;
         when Selection_Single | Selection_Browse =>
            if State xor Is_Selected (Widget, Index) then
               if State then
                  -- Deselect any selected
                  for Item in 1..Size loop
                     if Is_Selected (Widget, Item) then
                        Change_Selection (Widget, Item, False);
                        exit;
                     end if;
                  end loop;
                  Change_Selection (Widget, Index, True);
                  Changed := True;
               elsif Widget.Markup.Mode = Selection_Single then
                  Change_Selection (Widget, Index, False);
                  Changed := True;
               end if;
            end if;
         when Selection_Multiple =>
            if Is_Selected (Widget, Index) xor State then
               Change_Selection (Widget, Index, State);
               Changed := True;
            end if;
      end case;
      if Changed then
         Selection_Changed (Widget);
      end if;
   end Set_Selection;

   procedure Set_Selection
             (  Widget : access Gtk_Directory_Items_View_Record'Class;
                Name   : Item_Name;
                State  : Boolean
             )  is
      Position : constant Natural := Get_Index (Widget, Name);
   begin
      if Position > 0 then
         Set_Selection (Widget, Position, State);
      end if;
   end Set_Selection;

   procedure Set_Selection_Mode
             (  Widget : access Gtk_Directory_Items_View_Record'Class;
                Mode   : Gtk_Selection_Mode
             )  is
   begin
      if Widget.Markup.Mode = Mode then
         return;
      end if;
      declare
         Changed : Boolean := False;
         Size    : constant Natural := Get_Directory_Size (Widget);
         Current : Natural;
      begin
         Widget.Markup.Mode := Mode;
         case Mode is
            when Selection_None =>
               -- We just deselect anything selected
               for Index in 1..Size loop
                  if Is_Selected (Widget, Index) then
                     Change_Selection (Widget, Index, False);
                     Changed := True;
                  end if;
               end loop;
            when Selection_Single | Selection_Browse =>
               case Get_Selection_Size (Widget) is
                  when 0 =>
                     if Mode = Selection_Browse then
                        -- Nothing  selected.  When  there  is a current
                        -- item we select it, else we do the first item,
                        -- otherwise nothing.
                        Current := Get_Current (Widget);
                        if Current > 0 then
                           Change_Selection (Widget, Current, True);
                           Changed := True;
                        elsif Size > 0 then
                           Change_Selection (Widget, 1, True);
                           Changed := True;
                        end if;
                     end if;
                  when 1 =>
                     null;  -- One item selection, nothing to do
                  when others =>
                     -- Multiple selection, we shall deselect everything
                     -- but  one. Let's never deselect the current item,
                     -- otherwise it will be the first one that stays.
                     Current := Get_Current (Widget);
                     for Index in reverse 1..Size loop
                        if (  Current /= Index
                           and then
                              Is_Selected (Widget, Index)
                           )
                        then
                           Change_Selection (Widget, Index, False);
                           Changed := True;
                           exit when Get_Selection_Size (Widget) = 1;
                        end if;
                     end loop;
               end case;
            when Selection_Multiple =>
               -- There is nothing to do
               null;
         end case;
         if Changed then
            Selection_Changed (Widget);
         end if;
      end;
   end Set_Selection_Mode;

   procedure Set_Tree_Data
             (  Column : access Gtk_Tree_View_Column_Record'Class;
                Cell   : access Gtk_Cell_Renderer_Record'Class;
                Model  : access Gtk_Tree_Model_Record'Class;
                Iter   : Gtk_Tree_Iter;
                Data   : Gtk_Directory_Tree_View
             )  is
      Path : Gtk_Tree_Path := Get_Path (Model, Iter);
      Item : Gtk_Tree_Iter := To_Item (Data.Cache, Iter);
   begin
      Set_Property
      (  Data.Icon_Renderer,
         Stock_ID_Property,
         Get_Icon
         (  Data,
            Item_Type (Get_String (Data.Cache.Tree, Item, 0)),
            Row_Expanded (Data, Path),
            Has_Child (Data.Cache.Tree, Item)
      )  );
      Path_Free (Path);
      Set_Property
      (  Data.Name_Renderer,
         Text_Property,
         Get_String (Data.Cache.Tree, Item, 1)
      );
   end Set_Tree_Data;

   function To_Item
            (  Store     : access Gtk_Abstract_Directory_Record'Class;
               Directory : Gtk_Tree_Iter
            )  return Gtk_Tree_Iter is
      Item : Gtk_Tree_Iter;
   begin
      if Directory = Null_Iter then
         return Null_Iter;
      else
         Convert_Iter_To_Child_Iter (Store, Item, Directory);
         return Item;
      end if;
   end To_Item;

   function To_Marked
            (  Widget : access Gtk_Directory_Items_View_Record;
               Item   : Gtk_Tree_Iter
            )  return Gtk_Tree_Iter is
      Row : Gtk_Tree_Iter;
   begin
      if Item = Null_Iter then
         return Null_Iter;
      else
         Convert_Child_Iter_To_Iter (Widget.Content, Row, Item);
         return To_Extension (Widget.Markup, Row);
      end if;
   end To_Marked;

end Gtk.Abstract_Browser;
