--                                                                    --
--  procedure                       Copyright (c)  Dmitry A. Kazakov  --
--     Gtk.Directory_Browser                       Luebeck            --
--  Implementation                                 Autumn, 2007       --
--                                                                    --
--                                Last revision :  09:24 09 Apr 2010  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.Exceptions;             use Ada.Exceptions;
with Glib.Messages;              use Glib.Messages;
with GtkAda.Dialogs;             use GtkAda.Dialogs;
with Gtk.Enums;                  use Gtk.Enums;
with Gtk.Scrolled_Window;        use Gtk.Scrolled_Window;
with Gtk.Stock;                  use Gtk.Stock;
with Gtk.Tree_View;              use Gtk.Tree_View;
with Gtk.Widget;                 use Gtk.Widget;
with Strings_Edit.UTF8;          use Strings_Edit.UTF8;
with Strings_Edit.UTF8.Mapping;  use Strings_Edit.UTF8.Mapping;

with Ada.IO_Exceptions;
with Ada.Unchecked_Deallocation;

package body Gtk.Directory_Browser is

   type Case_Type is (Sensitive, Insensitive);
   Mode : Case_Type := Sensitive;

   function Where (Name : String) return String is
   begin
      return " in Gtk.Directory_Browser." & Name;
   end Where;

   procedure Say (Text : String) is
   begin
      if (  Button_OK
         =  Message_Dialog
            (  Dialog_Type   => Error,
               Buttons       => Button_OK,
               Title         => "File renaming error",
               Justification => Justify_Left,
               Msg           => Text
         )  )
      then
         null;
      end if;
   end Say;

   function Compare
            (  Store     : access Gtk_Directory_Record;
               Directory : Item_Path;
               A, B      : Directory_Item;
               By_Name   : Boolean
            )  return Row_Order is
   begin
      if not By_Name and then (A.Directory xor B.Directory) then
         if A.Directory then
            return Before;
         else
            return After;
         end if;
      end if;
      case Mode is
         when Sensitive =>
            if A.Name = B.Name then
               return Equal;
            elsif A.Name < B.Name then
               return Before;
            else
               return After;
            end if;
         when Insensitive =>
            declare
               A_Name  : UTF8_String := UTF8_String (A.Name);
               B_Name  : UTF8_String := UTF8_String (B.Name);
               A_Code  : UTF8_Code_Point;
               B_Code  : UTF8_Code_Point;
               A_Index : Integer := A.Name'First;
               B_Index : Integer := B.Name'First;
            begin
               while A_Index <= A.Name'Last loop
                  begin
                     Get (A_Name, A_Index, A_Code);
                  exception
                     when Ada.IO_Exceptions.Data_Error =>
                        exit;
                  end;
                  if B_Index > B.Name'Last then
                     return After;
                  end if;
                  begin
                     Get (B_Name, B_Index, B_Code);
                  exception
                     when Ada.IO_Exceptions.Data_Error =>
                        return After;
                  end;
                  A_Code := To_Lowercase (A_Code);
                  B_Code := To_Lowercase (B_Code);
                  if A_Code /= B_Code then
                     if A_Code < B_Code then
                        return Before;
                     else
                        return After;
                     end if;
                  end if;
               end loop;
               if B_Index > B.Name'Last then
                  return Equal;
               else
                  return Before;
               end if;
            end;
      end case;
   end Compare;

   procedure Delete
             (  Store : access Gtk_Directory_Record;
                File  : UTF8_String
             )  is
   begin
      case Kind (File) is
         when Directory =>
            Delete_Directory (File);
         when Ordinary_File | Special_File =>
            Delete_File (File);
      end case;
      Deleted (Store, Item_Path (File));
   end Delete;

   procedure Finalize (Reference : in out Browsing_Data_Ref) is
      procedure Free is
         new Ada.Unchecked_Deallocation
             (  Browsing_Data,
                Browsing_Data_Ptr
             );
      procedure Free is
         new Ada.Unchecked_Deallocation
             (  File_System_Root,
                File_System_Root_Ptr
             );
      Item : File_System_Root_Ptr := Reference.Ptr.Roots;
      Next : File_System_Root_Ptr := Reference.Ptr.Roots;
   begin
      while Item /= null loop
         Next := Item.Next;
         Free (Item);
         Item := Next;
      end loop;
      Free (Reference.Ptr);
   end Finalize;

   function Filter
            (  Widget    : access Gtk_Directory_Browser_Record;
               Directory : Boolean;
               Name      : Item_Name;
               Kind      : Item_Type
            )  return Boolean is
   begin
      return True;
   end Filter;

   function Filter
            (  Widget    : access Gtk_Files_View_Record;
               Directory : Boolean;
               Name      : Item_Name;
               Kind      : Item_Type
            )  return Boolean is
   begin
      return Filter (Widget.Browser, Directory, Name, Kind);
   end Filter;

   function Get_Cache
            (  Widget : access Gtk_Directory_Browser_Record
            )  return Gtk_Directory is
   begin
      return Widget.Cache;
   end Get_Cache;

   function Get_Directory
            (  Store : access Gtk_Directory_Record;
               Item  : Item_Path
            )  return Item_Path is
      Name   : constant UTF8_String := UTF8_String (Item);
      Result : constant UTF8_String := Get_Dirname (Name);
   begin
      if Result = "." or else Result = Name then
         raise Name_Error;
      else
         return Item_Path (Result);
      end if;
   end Get_Directory;

   function Get_Files_View
            (  Widget : access Gtk_Directory_Browser_Record
            )  return Gtk_Directory_Items_View is
   begin
      return Widget.Items.all'Access;
   end Get_Files_View;

   function Get_Name
            (  Store : access Gtk_Directory_Record;
               Item  : Item_Path
            )  return Item_Name is
      Name : constant UTF8_String := UTF8_String (Item);
   begin
      if Get_Root (Name) = Name then
         return Item_Name (Name);
      else
         return Item_Name (Get_Basename (Name));
      end if;
   exception
      when Use_Error =>
         return Item_Name (Name);
   end Get_Name;

   function Get_Path
            (  Store     : access Gtk_Directory_Record;
               Directory : Item_Path;
               Item      : Item_Name
            )  return Item_Path is
      Data : Browsing_Data renames Store.Data.Ptr.all;
   begin
      if Directory'Length = 0 then
         if Is_Root (Store, Item_Path (Item)) then
            return Item_Path (Item);
         else
            return
               Item_Path
               (  Build_Filename
                  (  Data.Roots.Name,
                     UTF8_String (Item)
               )  );
         end if;
      else
         return
            Item_Path
            (  Build_Filename
               (  UTF8_String (Directory),
                  UTF8_String (Item)
            )  );
      end if;
   end Get_Path;

   function Get_Tree_View
            (  Widget : access Gtk_Directory_Browser_Record
            )  return Gtk_Directory_Tree_View is
   begin
      return Widget.Tree.all'Access;
   end Get_Tree_View;

   function Get_Root_Directory (File : String) return String is
   begin
      return Get_Root (File);
   exception
      when Use_Error =>
         return File;
   end Get_Root_Directory;

   procedure Gtk_New
             (  Store  : out Gtk_Directory;
                Policy : Caching_Policy := Cache_Expanded
             )  is
   begin
      Store := new Gtk_Directory_Record;
      begin
         Initialize (Store, Policy);
      exception
         when Error : others =>
            Log
            (  GtkAda_Contributions_Domain,
               Log_Level_Critical,
               (  "Fault: "
               &  Exception_Information (Error)
               &  Where ("Gtk_New (Gtk_Directory)")
            )  );
            Unref (Store);
            Store := null;
            raise;
      end;
   end Gtk_New;

   procedure Gtk_New
             (  Widget    : out Gtk_Directory_Browser;
                File      : UTF8_String := Current_Directory;
                Columns   : Positive    := 4;
                Vertical  : Boolean     := False;
                Tree_Size : Gtk_Requisition :=
                               (Width => 180, Height => 500);
                List_Size : Gtk_Requisition :=
                               (Width => 600, Height => 500);
                Store     : Gtk_Directory := null
             )  is
   begin
      Widget := new Gtk_Directory_Browser_Record;
      begin
         Initialize
         (  Widget,
            File,
            Columns,
            Vertical,
            Tree_Size,
            List_Size,
            Store
         );
      exception
         when Error : others =>
            Log
            (  GtkAda_Contributions_Domain,
               Log_Level_Critical,
               (  "Fault: "
               &  Exception_Information (Error)
               &  Where ("Gtk_New (Gtk_Persistent_Storage_Tree_View)")
            )  );
            Destroy (Widget);
            Widget := null;
            raise;
      end;
   end Gtk_New;

   procedure Initialize
             (  Store  : access Gtk_Directory_Record'Class;
                Policy : Caching_Policy
             )  is
      procedure Insert (Root : String) is
      begin
         Store.Data.Ptr.Roots :=
            new File_System_Root'
                (  Length => Root'Length,
                   Next   => Store.Data.Ptr.Roots,
                   Name   => Root
                );
      end Insert;
   begin
      Store.Data.Ptr := new Browsing_Data;
      --
      -- Determining   the   existing   roots.   Unfortunately   neither
      -- Ada.Directories nor GLib does not provide anything better  than
      -- the following.  (GVolumeMonitor  just  does  not  work  in  the
      -- current version):
      --
      for Letter in reverse Character range 'A'..'Z' loop
         begin
            if Exists (Letter & ":\") then
               Insert (Letter & ":\");
               Mode := Insensitive;
            end if;
         exception
            when Name_Error =>
               null;
         end;
      end loop;
      if Store.Data.Ptr.Roots = null then
         begin
            if Exists ("/") then
               Insert ("/");
            end if;
         exception
            when Name_Error =>
               null;
         end;
      end if;
      Store.Data.Ptr.Policy := Policy;
      Store.Data.Ptr.Current_Root := Store.Data.Ptr.Roots;
      Gtk.Abstract_Browser.Initialize (Store);
   end Initialize;

   procedure Initialize
             (  Widget    : access Gtk_Directory_Browser_Record'Class;
                File      : UTF8_String;
                Columns   : Positive;
                Vertical  : Boolean;
                Tree_Size : Gtk_Requisition;
                List_Size : Gtk_Requisition;
                Store     : Gtk_Directory
             )  is
      Scroll : Gtk_Scrolled_Window;
   begin
      if Vertical then
         Initialize_Vpaned (Widget);
      else
         Initialize_Hpaned (Widget);
      end if;
      Widget.Cache := Store;
      if Widget.Cache = null then
         Gtk_New (Widget.Cache);
      end if;
      Gtk_New (Scroll);
      Set_Policy (Scroll, Policy_Automatic, Policy_Automatic);
      Add1 (Widget, Scroll);
      Widget.Tree := new Gtk_Folders_View_Record;
      Initialize (Widget.Tree, Widget.Cache, Item_Path (File));
      declare
         Size : Gtk_Requisition;
      begin
         Columns_Autosize (Widget.Tree);   -- Size columns
         Size_Request (Widget.Tree, Size); -- Query the integral size
         Set_Size_Request                  -- Set new size
         (  Widget.Tree,
            GInt'Max (1, GInt'Min (Size.Width,  Tree_Size.Width)),
            GInt'Max (1, GInt'Min (Size.Height, Tree_Size.Height))
         );
      end;
      Add (Scroll, Widget.Tree);
      Gtk_New (Scroll);
      Set_Policy (Scroll, Policy_Automatic, Policy_Automatic);
      Add2 (Widget, Scroll);
      Widget.Items := new Gtk_Files_View_Record;
      Widget.Items.Browser := Widget.all'Access;
      Initialize (Widget.Items, Widget.Tree, Columns);
      declare
         Size : Gtk_Requisition;
      begin
         Size_Request (Widget.Items, Size); -- Query the integral size
         Set_Size_Request                   -- Set new size
         (  Widget.Items,
            GInt'Max (1, GInt'Min (Size.Width,  List_Size.Width)),
            GInt'Max (1, GInt'Min (Size.Height, List_Size.Height))
         );
      end;
      Add (Scroll, Widget.Items);
      if Store = null then
         Unref (Widget.Cache);
      end if;
   exception
      when others =>
         if Store = null and then Widget.Cache /= null then
            Unref (Widget.Cache);
         end if;
         raise;
   end Initialize;

   function Is_Root
            (  Store : access Gtk_Directory_Record;
               Item  : Item_Path
            )  return Boolean is
      Name : constant UTF8_String := String (Item);
   begin
      return Is_Absolute (Name) and then Get_Root (Name) = Name;
   exception
      when Use_Error =>
         return False;
   end Is_Root;

   procedure Name_Commit
             (  Widget   : access Gtk_Folders_View_Record;
                Old_Path : Item_Path;
                New_Name : Item_Name
             )  is
      Cache : Gtk_Abstract_Directory renames Get_Cache (Widget);
   begin
      if Get_Name (Cache, Old_Path) /= New_Name then
         declare
            New_Path : Item_Path :=
               Get_Path
               (  Cache,
                  Get_Directory (Cache, Old_Path),
                  New_Name
               );
         begin
            Rename (String (Old_Path), String (New_Path));
         exception
            when Name_Error =>
               Say
               (  "Directory "
               &  String (Old_Path)
               &  " does not exist"
               );
               return;
            when Use_Error =>
               Say
               (  "Directory "
               &  String (Old_Path)
               &  " cannot be renamed to "
               &  String (New_Path)
               );
               return;
            when Error : others =>
               Say
               (  "Unable to rename "
               &  String (Old_Path)
               &  " to "
               &  String (New_Path)
               &  " due to internal error. "
               &  Exception_Message (Error)
               );
               return;
         end;
         Renamed (Cache, Old_Path, New_Name);
      end if;
   end Name_Commit;

   procedure Name_Commit
             (  Widget : access Gtk_Files_View_Record;
                Index  : Positive;
                Name   : Item_Name
             )  is
      Item    : Natural := Index;
      Changed : Boolean;
   begin
      if Get_Name (Widget, Index) /= Name then
         begin
            Rename
            (  String (Get_Path (Widget, Index)),
               String (Get_Path (Widget, Name))
            );
         exception
            when Name_Error =>
               Say
               (  "File "
               &  String (Get_Path (Widget, Index))
               &  " does not exist"
               );
               return;
            when Use_Error =>
               Say
               (  "File "
               &  String (Get_Path (Widget, Index))
               &  " cannot be renamed to "
               &  String (Get_Path (Widget, Name))
               );
               return;
            when Error : others =>
               Say
               (  "Unable to rename "
               &  String (Get_Path (Widget, Index))
               &  " to "
               &  String (Get_Path (Widget, Name))
               &  " due to internal error. "
               &  Exception_Message (Error)
               );
               return;
         end;
         Renamed (Widget, Item, Name);
         if Item > 0 then
            Move (Widget, Changed, 0, Item);
         end if;
      end if;
   end Name_Commit;

   function Read
            (  Store : access Gtk_Directory_Record
            )  return Directory_Item is
      This : Browsing_Data renames Store.Data.Ptr.all;
   begin
      case This.Mode is
         when None =>
            raise Ada.IO_Exceptions.End_Error;
         when Root =>
            if This.Current_Root = null then
               This.Mode := None;
               raise Ada.IO_Exceptions.End_Error;
            else
               if This.Current_Root.Name = "/" then
                  declare
                     Result : Directory_Item :=
                     (  Directory   => True,
                        Policy      => This.Policy,
                        Name_Length => This.Current_Root.Name'Length,
                        Kind_Length => Stock_Home'Length,
                        Kind => Item_Type (Stock_Home),
                        Name => Item_Name (This.Current_Root.Name)
                     );
                  begin
                     This.Current_Root := This.Current_Root.Next;
                     return Result;
                  end;
               else
                  declare
                     Result : Directory_Item :=
                     (  Directory   => True,
                        Policy      => This.Policy,
                        Name_Length => This.Current_Root.Name'Length,
                        Kind_Length => Stock_Harddisk'Length,
                        Kind => Item_Type (Stock_Harddisk),
                        Name => Item_Name (This.Current_Root.Name)
                     );
                  begin
                     This.Current_Root := This.Current_Root.Next;
                     return Result;
                  end;
               end if;
            end if;
         when Directory =>
            null;
      end case;
      begin
         while More_Entries (This.Search) loop
            Get_Next_Entry (This.Search, This.Current_Dir);
            declare
               Name : String renames Simple_Name (This.Current_Dir);
            begin
               if Name /= "." and then Name /= ".." then
                  case Kind (This.Current_Dir) is
                     when Directory =>
                        return
                        (  Directory   => True,
                           Policy      => This.Policy,
                           Name_Length => Name'Length,
                           Kind_Length => Stock_Directory'Length,
                           Name        => Item_Name (Name),
                           Kind        => Item_Type (Stock_Directory)
                        );
                     when Ordinary_File =>
                        return
                        (  Directory   => False,
                           Policy      => This.Policy,
                           Name_Length => Name'Length,
                           Kind_Length => Stock_File'Length,
                           Name        => Item_Name (Name),
                           Kind        => Item_Type (Stock_File)
                        );
                     when Special_File =>
                        return
                        (  Directory   => False,
                           Policy      => This.Policy,
                           Name_Length => Name'Length,
                           Kind_Length => Stock_Execute'Length,
                           Name        => Item_Name (Name),
                           Kind        => Item_Type (Stock_Execute)
                        );
                  end case;
              end if;
           end;
         end loop;
         End_Search (This.Search);
      exception
         when Error : others =>
           Raise_Exception
           (  Ada.IO_Exceptions.Data_Error'Identity,
              Exception_Message (Error)
           );
      end;
      This.Mode := None;
      raise Ada.IO_Exceptions.End_Error;
   end Read;

   function Rewind
            (  Store  : access Gtk_Directory_Record;
               Folder : Item_Path
            )  return Directory_Item is
      This : Browsing_Data renames Store.Data.Ptr.all;
   begin
      case This.Mode is
         when None | Root =>
            null;
         when Directory =>
            End_Search (This.Search);
      end case;
      if Folder'Length = 0 then
         This.Mode := Root;
         This.Current_Root := This.Roots;
      else
         This.Mode := None;
         Start_Search (This.Search, String (Folder), "");
         This.Mode := Directory;
      end if;
      return Directory_Entered;
   exception
      when Error : others =>
         Raise_Exception
         (  Ada.IO_Exceptions.Data_Error'Identity,
            Exception_Message (Error)
         );
   end Rewind;

end Gtk.Directory_Browser;
