with Ada.Exceptions;   use Ada.Exceptions;
with Ada.Text_IO;      use Ada.Text_IO;
with Ada.Command_Line; use Ada.Command_Line;

with Gdk.Threads;
with Gtk;
with Gtk.Main;
with Gtk.Window;
with Gtk.Button;
with Gtk.Label;
with Gtk.Enums;
with Gtk.Box;
with Gtk.Handlers;
with Gtkada.Intl;

procedure Locked_GUI is

   protected Counter_Monitor is
      procedure Start;
      entry Wait_For_Start;
      procedure Stop;
      function Check_If_Stop return Boolean;
      procedure Quit;
      function Check_If_Quit return Boolean;
      procedure Finished;
      function Check_If_Running return Boolean;
      entry Wait_Until_Finished;
   private
      Should_Start : Boolean := False;
      Should_Stop : Boolean := False;
      Should_Quit : Boolean := False;
      Is_Running : Boolean := False;
   end Counter_Monitor;

   protected body Counter_Monitor is
      procedure Start is
      begin
         Should_Start := True;
      end Start;

      entry Wait_For_Start
        when Should_Start or Should_Quit is
      begin
         Should_Start := False;
         Should_Stop := False;
         --  don't touch Should_Quit, so that it's persistant
         Is_Running := True;
      end Wait_For_Start;

      procedure Stop is
      begin
         Should_Stop := True;
      end Stop;

      function Check_If_Stop return Boolean is
      begin
         return Should_Stop;
      end Check_If_Stop;

      procedure Quit is
      begin
         Should_Quit := True;
      end Quit;

      function Check_If_Quit return Boolean is
      begin
         return Should_Quit;
      end Check_If_Quit;

      procedure Finished is
      begin
         Is_Running := False;
      end Finished;

      function Check_If_Running return Boolean is
      begin
         return Is_Running;
      end Check_If_Running;

      entry Wait_Until_Finished when not Is_Running is
      begin
         null;
      end Wait_Until_Finished;

   end Counter_Monitor;

   task Counter is
   end Counter;

   type Global_Window_Record is new Gtk.Window.Gtk_Window_Record with record
      Position_Vbox : Gtk.Box.Gtk_Vbox;
      Position_Hbox : Gtk.Box.Gtk_Hbox;
      Counter_Label : Gtk.Label.Gtk_Label;
      Run_Button    : Gtk.Button.Gtk_Button;
      Quit_Button   : Gtk.Button.Gtk_Button;
   end record;
   type Global_Window_Access is access Global_Window_Record'Class;
   Global_Window : Global_Window_Access;

   function "-" (Msg : String) return String is
      use Gtkada.Intl;
   begin
      return Dgettext ("hopetovo", Msg);
   end "-";

   procedure Quit is
   begin
      Counter_Monitor.Quit;
      --  BEWARE OF OGRES! May deadlock if lock is not given up!
      Gdk.Threads.Leave;
      Counter_Monitor.Wait_Until_Finished;
      Gdk.Threads.Enter;
      Destroy (Global_Window);
      Gtk.Main.Main_Quit;
   end Quit;

   procedure On_Quit_Button_Clicked
     (Button : access Gtk.Button.Gtk_Button_Record'Class)
   is
   begin
      pragma Unreferenced (Button);
      Quit;
   end On_Quit_Button_Clicked;

   procedure On_Run_Button_Clicked
     (Button : access Gtk.Button.Gtk_Button_Record'Class)
   is
   begin
      if Counter_Monitor.Check_If_Running then
         Button.all.Set_Sensitive (False);
         Counter_Monitor.Stop;
      else
         Counter_Monitor.Start;
      end if;
   end On_Run_Button_Clicked;

   procedure Init_GUI is
      package Button_Callback is new Gtk.Handlers.Callback (
         Gtk.Button.Gtk_Button_Record);
   begin
      Global_Window := new Global_Window_Record;
      Gtk.Window.Initialize (Global_Window, Gtk.Enums.Window_Toplevel);
      Set_Title (Global_Window, -"GUI Example");
      Gtk.Box.Gtk_New_Vbox (Global_Window.all.Position_Vbox, False, 40);
      Gtk.Box.Gtk_New_Hbox (Global_Window.all.Position_Hbox, True, 20);
      Gtk.Label.Gtk_New (Global_Window.all.Counter_Label, -"0");
      Gtk.Button.Gtk_New (Global_Window.all.Run_Button, -" Run ");
      Gtk.Button.Gtk_New (Global_Window.all.Quit_Button, -" Quit ");

      Button_Callback.Connect
        (Global_Window.all.Quit_Button,
         "clicked",
         Button_Callback.To_Marshaller (On_Quit_Button_Clicked'Access),
         False);
      --  XXX: window delete event should be also handled, but for simplicity
      --       reasons, it is ommitted.

      Button_Callback.Connect
        (Global_Window.all.Run_Button,
         "clicked",
         Button_Callback.To_Marshaller (On_Run_Button_Clicked'Access),
         False);

      Gtk.Box.Pack_Start
        (Global_Window.all.Position_Hbox,
         Global_Window.all.Run_Button,
         Expand  => True,
         Fill    => True,
         Padding => 20);
      Gtk.Box.Pack_Start
        (Global_Window.all.Position_Hbox,
         Global_Window.all.Quit_Button,
         Expand  => True,
         Fill    => True,
         Padding => 20);
      Gtk.Box.Pack_Start
        (Global_Window.all.Position_Vbox,
         Global_Window.all.Counter_Label,
         Expand  => True,
         Fill    => True,
         Padding => 40);
      Gtk.Box.Pack_Start
        (Global_Window.all.Position_Vbox,
         Global_Window.all.Position_Hbox,
         Expand  => True,
         Fill    => True,
         Padding => 0);
      Add (Global_Window, Global_Window.all.Position_Vbox);
      Show_All (Global_Window);

   end Init_GUI;

   task body Counter is
   begin
      Main_Loop :
      loop
         Counter_Monitor.Wait_For_Start;
         --  because the main thread is waiting for us on termination
         --  we can touch Gtk objects until we signal termination; beware of ogres, though
         --  more appropriate is exiting early
         exit Main_Loop when Counter_Monitor.Check_If_Quit;
         Gdk.Threads.Enter;
         --  Ada95 syntax
         Gtk.Button.Set_Label (Global_Window.all.Run_Button, -" Stop ");
         Gdk.Threads.Leave;
         Counter_Loop :
         for I in 0 .. 1000 loop
            exit Main_Loop when Counter_Monitor.Check_If_Quit;
            if Counter_Monitor.Check_If_Stop then
               exit Counter_Loop;
            end if;
            Gdk.Threads.Enter;
            Gtk.Label.Set_Label (Global_Window.all.Counter_Label, -(Integer'Image (I)));
            Gdk.Threads.Leave;
            delay 1.0;
         end loop Counter_Loop;
         Gdk.Threads.Enter;
         --  Ada2005 extended syntax
         Global_Window.all.Run_Button.all.Set_Label (-" Run ");
         Counter_Monitor.Finished;
         Global_Window.all.Run_Button.all.Set_Sensitive (True);
         Gdk.Threads.Leave;
      end loop Main_Loop;
      Counter_Monitor.Finished;
   end Counter;

begin
   Gtk.Main.Set_Locale;
   Gdk.Threads.G_Init;
   Gdk.Threads.Init;
   Gtk.Main.Init;
   Init_GUI;
   Gdk.Threads.Enter;
   Gtk.Main.Main;
   Gdk.Threads.Leave;
   return;

exception
   when An_Exception : others =>
      Put_Line (Exception_Information (An_Exception));
      Set_Exit_Status (Failure);
      raise;

end Locked_GUI;
